import warnings
from types import ModuleType
from typing import ClassVar, List

from marshmallow import INCLUDE, Schema, ValidationError, post_dump, post_load, validates, validates_schema

from bioimageio.spec.shared import fields
from . import raw_nodes
from .common import ValidationWarning


class SharedBioImageIOSchema(Schema):
    raw_nodes: ClassVar[ModuleType] = raw_nodes  # to be overwritten in subclass by version specific raw_nodes module
    short_bioimageio_description: ClassVar[str] = ""
    bioimageio_description: ClassVar[str] = ""

    @post_load
    def make_object(self, data, **kwargs):
        if data is None:
            return None

        this_type = getattr(self.raw_nodes, self.__class__.__name__, None)
        if this_type is None:
            # attempt import from shared raw nodes
            this_type = getattr(raw_nodes, self.__class__.__name__, None)
            if this_type is None:
                raise NotImplementedError(
                    f"neither {self.raw_nodes} nor {raw_nodes} has attribute {self.__class__.__name__}."
                )

        try:
            return this_type(**data)
        except TypeError as e:
            e.args += (f"when initializing {this_type} from {self}",)
            raise e

    def warn(self, field: str, msg: str):
        """warn about a field with a ValidationWarning"""
        simple_field_name = field.split("[")[0]  # field may include [idx]
        # field_instance = self.fields[simple_field_name]  # todo: account for <field>:<nested field>
        assert ": " not in field
        # todo: add spec trail to field
        # e.g. something similar to field = ":".join(self.context.get("field_path", []) + [field])
        # or: ":".join(field_instance.spec_trail)
        msg = f"{field}: {msg}"
        warnings.warn(msg, category=ValidationWarning)


class SharedProcessingSchema(Schema):
    """Used to generate Pre- and Postprocessing documentation.

    Define Pre-/Postprocessing schemas in the Preprocessing/Postprocessing schema that inherite from this class
    and they will be rendered in the documentation."""

    bioimageio_description: ClassVar[str]


class WithUnknown(SharedBioImageIOSchema):
    """allows to keep unknown fields on load and dump them the 'unknown' attribute of the data to serialize"""

    field_name_unknown_dict = "unknown"

    class Meta:
        unknown = INCLUDE

    @post_load
    def make_object(self, data, **kwargs):
        obj = super().make_object(data, **kwargs)
        assert hasattr(
            obj, self.field_name_unknown_dict
        ), f"expected raw node to have attribute {self.field_name_unknown_dict}"
        return obj

    @post_dump(pass_original=True)
    def keep_unknowns(self, output, orig, **kwargs):
        if orig and hasattr(orig, self.field_name_unknown_dict):
            out_w_unknown = fields.YamlDict()._serialize(
                getattr(orig, self.field_name_unknown_dict), self.field_name_unknown_dict, self
            )
            out_w_unknown.update(output)
            return out_w_unknown
        else:
            return output


class Dependencies(SharedBioImageIOSchema):
    manager = fields.String(bioimageio_description="Dependency manager For example: 'conda', 'maven', or 'pip'")
    file = fields.Union(
        [fields.URI(), fields.Path()],
        bioimageio_description="Dependency file. For example: 'environment.yaml', 'pom.xml', or 'requirements.txt'",
    )


class ParametrizedInputShape(SharedBioImageIOSchema):
    min = fields.List(
        fields.Integer(), required=True, bioimageio_description="The minimum input shape with same length as `axes`"
    )
    step = fields.List(
        fields.Integer(), required=True, bioimageio_description="The minimum shape change with same length as `axes`"
    )

    @validates_schema
    def matching_lengths(self, data, **kwargs):
        min_ = data["min"]
        step = data["step"]
        if min_ is None or step is None:
            return

        if len(min_) != len(step):
            raise ValidationError(f"'min' and 'step' have to have the same length! (min: {min_}, step: {step})")


class ImplicitOutputShape(SharedBioImageIOSchema):
    reference_tensor = fields.String(required=True, bioimageio_description="Name of the reference tensor.")
    scale = fields.List(
        fields.Float(), required=True, bioimageio_description="'output_pix/input_pix' for each dimension."
    )
    offset = fields.List(
        fields.Float(), required=True, bioimageio_description="Position of origin wrt to input. Multiple of 0.5."
    )

    @validates_schema
    def matching_lengths(self, data, **kwargs):
        scale = data["scale"]
        offset = data["offset"]
        if len(scale) != len(offset):
            raise ValidationError(f"scale {scale} has to have same length as offset {offset}!")

    @validates("offset")
    def double_offset_is_int(self, value: List[float]):
        for v in value:
            if 2 * v != int(2 * v):
                raise ValidationError(f"offset {v} in {value} not a multiple of 0.5!")



import sys
from argparse import ArgumentParser
from pathlib import Path

try:
    from ruamel.yaml import YAML

    ruamel_yaml = YAML(typ="safe")
except ImportError:
    raise RuntimeError("Cannot compare yaml syntax without the ruamel.yaml package")

try:
    import yaml as _pyyaml_yaml

    class pyyaml_yaml:
        @staticmethod
        def load(path: Path):
            with path.open() as f:
                return _pyyaml_yaml.load(f, None)  # fmt: off


except ImportError:
    raise RuntimeError("Cannot compare yaml syntax without the pyyaml package")


def parse_args():
    p = ArgumentParser(
        description="Check for differences between yaml 1.1 (using PyYAML) and yaml 1.2 syntax (using ruamel.yaml)."
    )
    p.add_argument(
        "--resource_description_path",
        type=Path,
        default=Path(__file__).parent / "../example_specs/models/unet2d_nuclei_broad/rdf.yaml",
    )
    args = p.parse_args()
    return args


def main(resource_description_path: Path):

    pyyaml = pyyaml_yaml.load(resource_description_path)
    assert isinstance(pyyaml, dict)
    ruamel = ruamel_yaml.load(resource_description_path)
    assert isinstance(ruamel, dict)

    diff = {key: (value, ruamel[key]) for key, value in pyyaml.items() if value != ruamel[key]}
    if diff:
        print(f"Found differences between yaml syntax 1.1/1.2 for {resource_description_path}:")
        print(diff)
    else:
        print(f"No differences found between yaml syntax 1.1/1.2 for {resource_description_path}:")

    return len(diff)


if __name__ == "__main__":
    args = parse_args()
    sys.exit(main(args.resource_description_path))



from bioimageio.spec.rdf.v0_2.schema import RDF
from bioimageio.spec.shared.schema import SharedBioImageIOSchema
from . import raw_nodes

try:
    from typing import get_args
except ImportError:
    from typing_extensions import get_args  # type: ignore


class _BioImageIOSchema(SharedBioImageIOSchema):
    raw_nodes = raw_nodes


class Dataset(_BioImageIOSchema, RDF):
    bioimageio_description = f"""# BioImage.IO Dataset Resource Description File Specification {get_args(raw_nodes.FormatVersion)[-1]}
This specification defines the fields used in a BioImage.IO-compliant resource description file (`RDF`) for describing datasets.
These fields are typically stored in a YAML file which we call Dataset Resource Description File or `dataset RDF`.

The dataset RDF YAML file contains mandatory and optional fields. In the following description, optional fields are indicated by _optional_.
_optional*_ with an asterisk indicates the field is optional depending on the value in another field.
"""



import copy
import pathlib
from typing import Any, Dict

from marshmallow import Schema

from . import raw_nodes, schema

AUTO_CONVERTED_DOCUMENTATION_FILE_NAME = "auto_converted_documentation.md"


def convert_model_v0_3_1_to_v0_3_2(data: Dict[str, Any]) -> Dict[str, Any]:
    data["type"] = "model"
    data["format_version"] = "0.3.2"
    future = data.get("config", {}).get("future", {}).pop("0.3.2", {})

    authors = data.get("authors")
    if isinstance(authors, list):
        data["authors"] = [{"name": name} for name in authors]
        authors_update = future.get("authors")
        if authors_update is not None:
            for a, u in zip(data["authors"], authors_update):
                a.update(u)

    # packaged_by
    packaged_by = data.get("packaged_by")
    if packaged_by is not None:
        data["packaged_by"] = [{"name": name} for name in data["packaged_by"]]
        packaged_by_update = future.get("packaged_by")
        if packaged_by_update is not None:
            for a, u in zip(data["packaged_by"], packaged_by_update):
                a.update(u)

    # authors of weights
    weights = data.get("weights")
    if isinstance(weights, dict):
        for weights_format, weights_entry in weights.items():
            if "authors" not in weights_entry:
                continue

            weights_entry["authors"] = [{"name": name} for name in weights_entry["authors"]]
            authors_update = future.get("weights", {}).get(weights_format, {}).get("authors")
            if authors_update is not None:
                for a, u in zip(weights_entry["authors"], authors_update):
                    a.update(u)

    # documentation: we now enforce `documentation` to be a local md file
    if "documentation" in data:

        class DocSchema(Schema):
            doc = schema.Model().fields["documentation"]

        doc_errors = DocSchema().validate({"doc": data["documentation"]})
        if doc_errors:
            # data["documentation"] is not a local relative md file, so we replace it with a placeholder.
            # Having access only to the raw data dict, we cannot write the AUTO_CONVERTED_DOCUMENTATION_FILE_NAME file, but
            # save the original content of data["documentation"] in data["config"][AUTO_CONVERTED_DOCUMENTATION_FILE_NAME]
            # to be written to AUTO_CONVERTED_DOCUMENTATION_FILE_NAME at a later stage.
            data["config"] = data.get("config", {})  # make sure config exists
            if AUTO_CONVERTED_DOCUMENTATION_FILE_NAME not in data["config"]:
                orig_doc = data["documentation"]
                assert isinstance(orig_doc, str)
                if orig_doc.startswith("http"):
                    if orig_doc.endswith(".md"):
                        doc = raw_nodes.URI(orig_doc)
                    else:
                        doc = f"Find documentation at {orig_doc}"
                else:
                    doc = pathlib.Path(orig_doc)

                data["config"][AUTO_CONVERTED_DOCUMENTATION_FILE_NAME] = doc
                data["documentation"] = AUTO_CONVERTED_DOCUMENTATION_FILE_NAME

    # model version
    if "version" in future:
        data["version"] = future.pop("version")

    return data


def convert_model_v0_3_2_to_v0_3_3(data: Dict[str, Any]) -> Dict[str, Any]:
    data["format_version"] = "0.3.3"
    if "outputs" in data:
        for out in data["outputs"]:
            if "shape" in out:
                shape = out["shape"]
                if isinstance(shape, dict) and "reference_input" in shape:
                    shape["reference_tensor"] = shape.pop("reference_input")

    return data


def maybe_convert(data: Dict[str, Any]) -> Dict[str, Any]:
    """auto converts model 'data' to newest format"""

    data = copy.deepcopy(data)

    if data.get("format_version", "0.3.0") == "0.3.0":
        # no breaking change, bump to 0.3.1
        data["format_version"] = "0.3.1"

    if data["format_version"] == "0.3.1":
        data = convert_model_v0_3_1_to_v0_3_2(data)

    if data["format_version"] == "0.3.2":
        data = convert_model_v0_3_2_to_v0_3_3(data)

    if data["format_version"] in ("0.3.3", "0.3.4", "0.3.5"):
        data["format_version"] = "0.3.6"

    # remove 'future' from config if no other than the used future entries exist
    config = data.get("config", {})
    if config.get("future") == {}:
        del config["future"]

    # remove 'config' if now empty
    if data.get("config") == {}:
        del data["config"]

    return data



import typing

from marshmallow.validate import (
    ContainsNoneOf,
    Equal,
    Length,
    OneOf,
    Predicate as MarshmallowPredicate,
    Range,
    URL as MarshmallowURL,
    ValidationError,
    Validator,
)

ContainsNoneOf = ContainsNoneOf
Equal = Equal
Length = Length
OneOf = OneOf
Range = Range


class Attribute(Validator):

    default_message = "Invalid input."

    def __init__(
        self,
        attribute: str,
        validate: typing.Union[
            typing.Callable[[typing.Any], typing.Any], typing.Iterable[typing.Callable[[typing.Any], typing.Any]]
        ],
        is_getter_method: bool = False,
    ):
        self.attribute = attribute
        self.is_getter_method = is_getter_method
        self.validate = [validate] if callable(validate) else validate

    def _repr_args(self) -> str:
        return "attribute={!r}, is_getter_method={!r}, validate={!r}".format(
            self.attribute, self.is_getter_method, self.validate
        )

    def __call__(self, value: typing.Any) -> typing.Any:
        try:
            attr = getattr(value, self.attribute)
            if self.is_getter_method:
                attr = attr()
        except Exception as e:
            raise ValidationError(str(e)) from e

        try:
            return all(validator(attr) for validator in self.validate)
        except Exception as e:
            raise ValidationError(f"Invalid {self.attribute} ({value}): {str(e)}") from e


class Predicate(MarshmallowPredicate):
    """extends marshmallow.Predicate by accepting *args and 'invert_output' .
    Call the specified ``method`` of the ``value`` object. The
    validator succeeds if the invoked method returns an object that
    evaluates to True in a Boolean context. Any additional arguments
    and keyword arguments will be passed to the method.

    :param method: The name of the method to invoke.
    :param args: Additional arguments to pass to the method.
    :param invert_output: Flag to succeed if method returns an object that evaluates to False instead of True.
    :param error: Error message to raise in case of a validation error.
        Can be interpolated with `{input}` and `{method}`.
    :param kwargs: Additional keyword arguments to pass to the method.
    """

    default_message = "Invalid input."

    def __init__(self, method: str, *args, invert_output: bool = False, error: typing.Optional[str] = None, **kwargs):
        super().__init__(method, error=error, **kwargs)
        self.args = args
        self.invert_output = invert_output

    def _repr_args(self) -> str:
        return "method={!r}, invert_output={!r}, args={!r}, kwargs={!r}".format(
            self.method, self.invert_output, self.args, self.kwargs
        )

    def __call__(self, value: typing.Any) -> typing.Any:
        method = getattr(value, self.method)

        ret = method(*self.args, **self.kwargs)
        if self.invert_output:
            ret = not ret

        if not ret:
            raise ValidationError(self._format_error(value))

        return value


class URL(MarshmallowURL):
    def __call__(self, value: typing.Any):
        return super().__call__(str(value))  # cast value which might be a raw_nodes.URI to string



""" raw nodes for the collection RDF spec

raw nodes are the deserialized equivalent to the content of any RDF.
serialization and deserialization are defined in schema:
RDF <--schema--> raw nodes
"""
import pathlib
from dataclasses import dataclass
from pathlib import Path
from typing import Any, Dict, List, Union

import packaging.version
from marshmallow import missing
from marshmallow.utils import _Missing

from bioimageio.spec.rdf.v0_2.raw_nodes import Author, Badge, CiteEntry, Maintainer, RDF
from bioimageio.spec.shared.raw_nodes import RawNode, URI

try:
    from typing import Literal, get_args
except ImportError:
    from typing_extensions import Literal, get_args  # type: ignore

FormatVersion = Literal[
    "0.2.0", "0.2.1", "0.2.2"
]  # newest format needs to be last (used to determine latest format version)


@dataclass
class CollectionEntry(RawNode):
    rdf_source: Union[_Missing, URI] = missing
    rdf_update: Dict[str, Any] = missing

    def __init__(
        self, rdf_source: Union[_Missing, URI] = missing, rdf_update: Dict[str, Any] = missing, **implicit_rdf_update
    ):
        self.rdf_source = rdf_source
        self.rdf_update = rdf_update or {}
        self.rdf_update.update(implicit_rdf_update)
        super().__init__()


@dataclass
class Collection(RDF):
    collection: List[CollectionEntry] = missing
    unknown: Dict[str, Any] = missing
    # manual __init__ to allow for unknown kwargs
    def __init__(
        self,
        *,
        # ResourceDescription
        format_version: FormatVersion,
        name: str,
        type: str = missing,
        version: Union[_Missing, packaging.version.Version] = missing,
        root_path: pathlib.Path = pathlib.Path(),
        # RDF
        attachments: Union[_Missing, Dict[str, Any]] = missing,
        authors: Union[_Missing, List[Author]] = missing,
        badges: Union[_Missing, List[Badge]] = missing,
        cite: Union[_Missing, List[CiteEntry]] = missing,
        config: Union[_Missing, dict] = missing,
        covers: Union[_Missing, List[Union[URI, Path]]] = missing,
        description: str,
        documentation: Union[_Missing, Path] = missing,
        git_repo: Union[_Missing, str] = missing,
        license: Union[_Missing, str] = missing,
        links: Union[_Missing, List[str]] = missing,
        maintainers: Union[_Missing, List[Maintainer]] = missing,
        tags: Union[_Missing, List[str]] = missing,
        # collection RDF
        collection: List[CollectionEntry],
        unknown: Dict[str, Any] = missing,
        **implicitly_unknown,
    ):
        self.collection = collection
        self.unknown = unknown or {}
        self.unknown.update(implicitly_unknown)
        super().__init__(
            attachments=attachments,
            authors=authors,
            badges=badges,
            cite=cite,
            config=config,
            covers=covers,
            description=description,
            documentation=documentation,
            format_version=format_version,
            git_repo=git_repo,
            license=license,
            links=links,
            maintainers=maintainers,
            name=name,
            tags=tags,
            type=type,
            version=version,
            root_path=root_path,
        )



import torch
import torch.nn as nn


class Upsample(nn.Module):
    def __init__(self, scale_factor, mode="bilinear"):
        super().__init__()
        self.scale_factor = scale_factor
        self.mode = mode

    def forward(self, input):
        return nn.functional.interpolate(input, scale_factor=self.scale_factor, mode=self.mode, align_corners=False)


class UNet2d(nn.Module):
    def __init__(self, input_channels, output_channels, training=False):
        super().__init__()
        self.input_channels = input_channels
        self.output_channels = output_channels
        self.n_levels = 3

        self.encoders = nn.ModuleList(
            [self.conv_layer(self.input_channels, 16), self.conv_layer(16, 32), self.conv_layer(32, 64)]
        )
        self.downsamplers = nn.ModuleList([self.downsampler()] * self.n_levels)

        self.base = self.conv_layer(64, 128)

        self.decoders = nn.ModuleList([self.conv_layer(128, 64), self.conv_layer(64, 32), self.conv_layer(32, 16)])
        self.upsamplers = nn.ModuleList([self.upsampler(128, 64), self.upsampler(64, 32), self.upsampler(32, 16)])

        self.output = nn.Conv2d(16, self.output_channels, 1)
        self.training = training

    def conv_layer(self, in_channels, out_channels):
        kernel_size = 3
        padding = 1
        return nn.Sequential(
            nn.Conv2d(in_channels, out_channels, kernel_size, padding=padding),
            nn.Conv2d(out_channels, out_channels, kernel_size, padding=padding),
            nn.ReLU(inplace=True),
        )

    def downsampler(self):
        return nn.MaxPool2d(2)

    def upsampler(self, in_channels, out_channels):
        return nn.Sequential(Upsample(2), nn.Conv2d(in_channels, out_channels, 1))

    def forward(self, input):
        x = input

        from_encoder = []
        for encoder, sampler in zip(self.encoders, self.downsamplers):
            x = encoder(x)
            from_encoder.append(x)
            x = sampler(x)

        x = self.base(x)

        for decoder, sampler, enc in zip(self.decoders, self.upsamplers, from_encoder[::-1]):
            x = sampler(x)
            x = torch.cat([enc, x], dim=1)
            x = decoder(x)

        x = self.output(x)

        # apply a sigmoid directly if we are in inference mode
        if not self.training:
            # postprocessing
            x = torch.sigmoid(x)

        return x



import os
import traceback
import warnings
from copy import deepcopy
from pathlib import Path
from typing import Any, Dict, IO, List, Optional, Union

from marshmallow import ValidationError

from .collection.v0_2.utils import resolve_collection_entries
from .io_ import (
    load_raw_resource_description,
    resolve_rdf_source,
    save_raw_resource_description,
    serialize_raw_resource_description_to_dict,
)
from .shared import update_nested
from .shared.common import ValidationWarning, nested_default_dict_as_nested_dict, yaml
from .shared.raw_nodes import ResourceDescription as RawResourceDescription
from .v import __version__

try:
    from typing import TypedDict, Literal
except ImportError:
    from typing_extensions import TypedDict, Literal  # type: ignore


def update_format(
    rdf_source: Union[dict, os.PathLike, IO, str, bytes],
    path: Union[os.PathLike, str],
    update_to_format: str = "latest",
):
    """Update a BioImage.IO resource"""
    raw = load_raw_resource_description(rdf_source, update_to_format=update_to_format)
    save_raw_resource_description(raw, Path(path))


class ValidationSummary(TypedDict):
    bioimageio_spec_version: str
    error: Union[None, str, Dict[str, Any]]
    name: str
    nested_errors: Dict[str, dict]
    source_name: str
    status: Union[Literal["passed", "failed"], str]
    traceback: Optional[List[str]]
    warnings: dict


def validate(
    rdf_source: Union[RawResourceDescription, dict, os.PathLike, IO, str, bytes],
    update_format: bool = False,
    update_format_inner: bool = None,
    verbose: bool = "deprecated",  # type: ignore
) -> ValidationSummary:
    """Validate a BioImage.IO Resource Description File (RDF).

    Args:
        rdf_source: resource description as path, url or bytes of an RDF or packaged resource, or as yaml string or dict
        update_format: weather or not to apply auto-conversion to the latest format version before validation
        update_format_inner: (applicable to `collections` resources only) `update_format` for nested resources
        verbose: deprecated

    Returns:
        A summary dict with keys:
            bioimageio_spec_version,
            error,
            name,
            nested_errors,
            source_name,
            status,
            traceback,
            warnings,
    """
    if verbose != "deprecated":
        warnings.warn("'verbose' flag is deprecated")

    if update_format_inner is None:
        update_format_inner = update_format

    error: Union[None, str, Dict[str, Any]] = None
    tb = None
    nested_errors: Dict[str, dict] = {}
    if isinstance(rdf_source, RawResourceDescription):
        source_name = rdf_source.name
    else:
        try:
            rdf_source_preview, source_name, root = resolve_rdf_source(rdf_source)
        except Exception as e:
            error = str(e)
            tb = traceback.format_tb(e.__traceback__)
            try:
                source_name = str(rdf_source)
            except Exception as e:
                source_name = str(e)
        else:
            if not isinstance(rdf_source_preview, dict):
                error = f"expected loaded resource to be a dictionary, but got type {type(rdf_source_preview)}"

    raw_rd = None
    format_version = ""
    resource_type = ""
    if error:
        validation_warnings = []
    else:
        with warnings.catch_warnings(record=True) as all_warnings:
            try:
                raw_rd = load_raw_resource_description(rdf_source, update_to_format="latest" if update_format else None)
            except ValidationError as e:
                error = nested_default_dict_as_nested_dict(e.normalized_messages())
            except Exception as e:
                error = str(e)
                tb = traceback.format_tb(e.__traceback__)

            if raw_rd is not None:
                format_version = raw_rd.format_version
                resource_type = "general" if raw_rd.type == "rdf" else raw_rd.type

            if raw_rd is not None and raw_rd.type == "collection":
                assert hasattr(raw_rd, "collection")
                for idx, (entry_rdf, entry_error) in enumerate(resolve_collection_entries(raw_rd)):  # type: ignore
                    if entry_error:
                        entry_summary: Union[Dict[str, str], ValidationSummary] = {"error": entry_error}
                    else:
                        assert isinstance(entry_rdf, RawResourceDescription)
                        entry_summary = validate(
                            entry_rdf, update_format=update_format, update_format_inner=update_format_inner
                        )

                        wrns: Union[str, dict] = entry_summary.get("warnings", {})
                        assert isinstance(wrns, dict)
                        id_info = f"(id={entry_rdf.id}) " if hasattr(entry_rdf, "id") else ""  # type: ignore
                        for k, v in wrns.items():
                            warnings.warn(f"collection[{idx}]:{k}: {id_info}{v}", category=ValidationWarning)

                    if entry_summary["error"]:
                        if "collection" not in nested_errors:
                            nested_errors["collection"] = {}

                        nested_errors["collection"][idx] = entry_summary["error"]

                if nested_errors:
                    # todo: make short error message and refer to 'nested_errors' or deprecated 'nested_errors'
                    error = nested_errors

        validation_warnings = [w for w in all_warnings if issubclass(w.category, ValidationWarning)]

    return {
        "bioimageio_spec_version": __version__,
        "error": error,
        "name": (
            f"bioimageio.spec static validation of {resource_type} RDF {format_version}"
            f"{' with update to latest format version' if update_format else ''}"
        ),
        "nested_errors": nested_errors,
        "source_name": source_name,
        "status": "passed" if error is None else "failed",
        "traceback": tb,
        "warnings": ValidationWarning.get_warning_summary(validation_warnings),
    }


def update_rdf(
    source: Union[RawResourceDescription, dict, os.PathLike, IO, str, bytes],
    update: Union[RawResourceDescription, dict, os.PathLike, IO, str, bytes],
    output: Union[dict, os.PathLike] = None,
    validate_output: bool = True,
) -> Union[dict, Path, RawResourceDescription]:
    """
    Args:
        source:  source of RDF
        update:  a (partial) RDF used as update
        output:  dict or path to write output to (default: return new dict)
        validate_output: whether or not to validate the updated RDF

    Returns:
        The updated content of the source rdf as dict or,
        if output is a path, that path (where the updated content is saved to).

    Raises:
        ValidationError: if `validate_output` and the updated rdf does not pass validation
    """
    if isinstance(source, RawResourceDescription):
        src = source
    else:
        src = load_raw_resource_description(source)

    up = resolve_rdf_source(update)

    if src.root_path != up.root and not validate_output:
        warnings.warn(
            f"root path of source {src.name} and update {up.name} differ. Relative paths might be invalid in the output."
        )

    out_data = update_nested(src, up.data)
    assert isinstance(out_data, (RawResourceDescription, dict))
    if validate_output:
        summary = validate(out_data)
        if summary["warnings"]:
            warnings.warn(f"updated rdf validation warnings\n: {summary['warnings']}")
        if summary["status"] != "passed":
            msg = f"updated rdf did not pass validation; status: {summary['status']}"
            if summary["error"]:
                msg += f"; error: {summary['error']}"

            raise ValidationError(msg)

    if output is None:
        if isinstance(source, RawResourceDescription):
            return load_raw_resource_description(out_data)
        else:
            output = {}

    if isinstance(output, dict):
        if isinstance(out_data, RawResourceDescription):
            out_data = serialize_raw_resource_description_to_dict(out_data, convert_absolute_paths=False)

        assert isinstance(out_data, dict)
        output.update(out_data)
        return output
    else:
        assert yaml is not None
        output = Path(output)
        if isinstance(out_data, RawResourceDescription):
            out_data.root_path = output.parent
            try:
                out_data = serialize_raw_resource_description_to_dict(out_data, convert_absolute_paths=True)
            except ValueError as e:
                warnings.warn(
                    f"Failed to convert paths in updated rdf to relative paths with root {output}; error: {e}"
                )
                warnings.warn(f"updated rdf at {output} contains absolute paths and is thus invalid!")
                out_data = serialize_raw_resource_description_to_dict(out_data, convert_absolute_paths=False)

        yaml.dump(out_data, output)
        return output



from typing import TypeVar, Union

from .node_transformer import NestedUpdateTransformer
from .raw_nodes import RawNode

UpdateType = TypeVar("UpdateType")


def update_nested(data: Union[dict, list, RawNode], update: UpdateType) -> Union[dict, list, RawNode, UpdateType]:
    return NestedUpdateTransformer().transform(data, update)



import os
import pathlib
from typing import List, Optional, Tuple, Union

from marshmallow import missing
from marshmallow.utils import _Missing

from . import raw_nodes, schema
from bioimageio.spec.shared.raw_nodes import ResourceDescription as RawResourceDescription


def filter_resource_description(raw_rd: raw_nodes.RDF) -> raw_nodes.RDF:
    return raw_rd


def resolve_collection_entries(
    collection: raw_nodes.Collection, collection_id: Optional[str] = None, update_to_format: Optional[str] = None
) -> List[Tuple[Optional[RawResourceDescription], Optional[str]]]:
    from bioimageio.spec import serialize_raw_resource_description_to_dict, load_raw_resource_description

    ret = []
    seen_ids = set()
    for idx, entry in enumerate(collection.collection):  # type: ignore
        entry_error: Optional[str] = None
        id_info = f"(id={entry.rdf_update['id']}) " if "id" in entry.rdf_update else ""

        # rdf entries are based on collection RDF...
        rdf_data = serialize_raw_resource_description_to_dict(collection)
        assert missing not in rdf_data.values()
        rdf_data.pop("collection")  # ... without the collection field to avoid recursion

        root_id = rdf_data.pop("id", None) if collection_id is None else collection_id
        # update rdf entry with entry's rdf_source
        sub_id: Union[str, _Missing] = missing
        if entry.rdf_source is not missing:
            rdf_source = entry.rdf_source
            if isinstance(rdf_source, str) and not rdf_source.startswith("http") or isinstance(rdf_source, os.PathLike):
                # a relative rdf_source path is relative to collection.root_path
                rdf_source = collection.root_path / pathlib.Path(rdf_source)

            try:
                source_entry_rd = load_raw_resource_description(rdf_source)
            except Exception as e:
                entry_error = f"collection[{idx}]: {id_info}Invalid rdf_source: {e}"
            else:
                source_entry_data = serialize_raw_resource_description_to_dict(source_entry_rd)
                sub_id = source_entry_data.pop("id", missing)
                assert missing not in source_entry_data.values()
                rdf_data.update(source_entry_data)

        # update rdf entry with fields specified directly in the entry
        rdf_update = schema.CollectionEntry().dump(entry)
        assert missing not in rdf_update.values()
        sub_id = rdf_update.pop("id", sub_id)
        if sub_id is missing:
            entry_error = f"collection[{idx}]: Missing `id` field"
        elif sub_id in seen_ids:
            entry_error = f"collection[{idx}]: Duplicate `id` value {sub_id}"
        else:
            seen_ids.add(sub_id)

        rdf = None
        if entry_error is None:
            rdf_data.update(rdf_update)
            if root_id is None:
                rdf_data["id"] = sub_id
            else:
                rdf_data["id"] = f"{root_id}/{sub_id}"

            # Convert simple links to links with collection id prepended
            if "links" in rdf_data:
                for i in range(len(rdf_data["links"])):
                    link = rdf_data["links"][i]
                    if "/" not in link:
                        rdf_data["links"][i] = collection.id + "/" + link

            rdf_data.pop("rdf_source", None)  # remove absorbed rdf_source
            rdf_data["root_path"] = collection.root_path  # collection entry always has the same root as the collection
            assert missing not in rdf_data.values()
            try:
                rdf = load_raw_resource_description(rdf_data, update_to_format=update_to_format)
            except Exception as e:
                entry_error = str(e)

        ret.append((rdf, entry_error))

    return ret



""" raw nodes for the dataset RDF spec

raw nodes are the deserialized equivalent to the content of any RDF.
serialization and deserialization are defined in schema:
RDF <--schema--> raw nodes
"""
from dataclasses import dataclass

from marshmallow import missing

from bioimageio.spec.rdf.v0_2.raw_nodes import FormatVersion, RDF as _RDF

try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal  # type: ignore

FormatVersion = FormatVersion


@dataclass
class Dataset(_RDF):
    type: Literal["dataset"] = missing



import dataclasses
import os
import pathlib
import typing

from marshmallow import missing
from marshmallow.utils import _Missing

from . import raw_nodes
from ._resolve_source import resolve_local_source, resolve_source as _resolve_source

try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal  # type: ignore

GenericResolvedNode = typing.TypeVar("GenericResolvedNode", bound=raw_nodes.RawNode)
GenericRawNode = typing.TypeVar("GenericRawNode", bound=raw_nodes.RawNode)
GenericRawRD = typing.TypeVar("GenericRawRD", bound=raw_nodes.ResourceDescription)
URI_Type = typing.TypeVar("URI_Type", bound=raw_nodes.URI)


def iter_fields(node: GenericRawNode):
    for field in dataclasses.fields(node):
        yield field.name, getattr(node, field.name)


class NodeVisitor:
    def visit(self, node: typing.Any) -> None:
        method = "visit_" + node.__class__.__name__

        visitor: typing.Callable[[typing.Any], typing.Any] = getattr(self, method, self.generic_visit)

        visitor(node)

    def generic_visit(self, node):
        """Called if no explicit visitor function exists for a node."""

        if isinstance(node, raw_nodes.RawNode):
            for field, value in iter_fields(node):
                self.visit(value)
        elif isinstance(node, dict):
            [self.visit(subnode) for subnode in node.values()]
        elif isinstance(node, (tuple, list)):
            [self.visit(subnode) for subnode in node]


class Transformer:
    def transform(self, node: typing.Any, **kwargs) -> typing.Any:
        method = "transform_" + node.__class__.__name__

        transformer = getattr(self, method, self.generic_transformer)

        return transformer(node, **kwargs)  # noqa

    def generic_transformer(self, node: typing.Any, **kwargs) -> typing.Any:
        return node

    def transform_list(self, node: list, **kwargs) -> list:
        return [self.transform(subnode, **kwargs) for subnode in node]

    def transform_dict(self, node: dict, **kwargs) -> dict:
        return {key: self.transform(value, **kwargs) for key, value in node.items()}


class NestedUpdateTransformer:
    """update a nested dict/list/raw_node with a nested dict/list update"""

    DROP = "DROP"
    KEEP = "KEEP"

    def transform(self, node: typing.Any, update: typing.Any) -> typing.Any:
        if update == self.KEEP:
            return node

        if isinstance(update, raw_nodes.RawNode):
            raise TypeError("updating with raw node is not allowed")

        method = "transform_" + node.__class__.__name__
        transformer = getattr(self, method, self.generic_transformer)

        return transformer(node, update)  # noqa

    def generic_transformer(self, node: typing.Any, update: typing.Any) -> typing.Any:
        if isinstance(node, raw_nodes.RawNode):
            return self.transform_node(node, update)
        else:
            return update

    def transform_node(
        self, node: raw_nodes.RawNode, update: typing.Union[dict, typing.Any]
    ) -> typing.Union[raw_nodes.RawNode, typing.Any]:
        if isinstance(update, dict):
            updated_kwargs = {
                name: self.transform(value, update.get(name, self.KEEP)) for name, value in iter_fields(node)
            }

            if "format_version" in update:
                # add new fields
                for k in set(update) - set(updated_kwargs):
                    updated_kwargs[k] = update[k]

                # todo: resolve raw node for updated format_version
                raise NotImplementedError("Updating format_version not yet implemented")
            else:
                invalid_updates = set(update) - set(updated_kwargs)
                if invalid_updates:
                    raise ValueError(f"Got unexpected updates for non-existing fields: {invalid_updates}")

            return dataclasses.replace(node, **updated_kwargs)
        else:
            return update

    def transform_list(self, node: list, update: typing.Union[list, typing.Any]) -> typing.Union[list, typing.Any]:
        if isinstance(update, list):
            if len(update) < len(node):
                update = update + [self.KEEP] * (len(node) - len(update))

            if len(node) < len(update):
                node = node + [self.DROP] * (len(update) - len(node))

            node = [self.transform(n, u) for n, u in zip(node, update)]
            return [e for e in node if e != self.DROP]
        else:
            return update

    def transform_dict(self, node: dict, update: typing.Union[dict, typing.Any]) -> typing.Union[dict, typing.Any]:
        if isinstance(update, dict):
            ret = {k: self.transform(v, update.get(k, self.KEEP)) for k, v in node.items()}
            for k, v in update.items():
                if k not in ret:
                    ret[k] = v

            return {k: v for k, v in ret.items() if v != self.DROP}
        else:
            return update


class NodeTransformer(Transformer):
    def generic_transformer(self, node: GenericRawNode, **kwargs) -> GenericRawNode:
        if isinstance(node, raw_nodes.RawNode):
            return dataclasses.replace(
                node, **{name: self.transform(value, **kwargs) for name, value in iter_fields(node)}
            )
        else:
            return super().generic_transformer(node, **kwargs)


class NodeTransformerKnownParent(NodeTransformer):
    def generic_transformer(
        self,
        node: GenericRawNode,
        name: typing.Optional[str] = None,
        parent: typing.Optional[raw_nodes.RawNode] = None,
        **kwargs,
    ) -> GenericRawNode:
        if isinstance(node, raw_nodes.RawNode):
            return dataclasses.replace(
                node, **{n: self.transform(value, name=n, parent=node) for n, value in iter_fields(node)}
            )
        else:
            return super().generic_transformer(node, name=name, parent=parent)


class PathToRemoteUriTransformer(NodeTransformer):
    def __init__(self, *, remote_source: URI_Type):
        remote_path = pathlib.PurePosixPath(remote_source.path).parent.as_posix()
        self.remote_root = dataclasses.replace(remote_source, path=remote_path, uri_string=None)

    def transform_URI(self, node: URI_Type) -> URI_Type:
        if node.scheme == "file":
            assert not node.authority
            assert not node.query
            assert not node.fragment
            return self._transform_Path(pathlib.Path(node.path))

        return node

    def _transform_Path(self, leaf: pathlib.PurePath):
        assert not leaf.is_absolute()
        path = pathlib.PurePosixPath(self.remote_root.path) / leaf
        return dataclasses.replace(self.remote_root, path=path.as_posix(), uri_string=None)

    def transform_PurePath(self, leaf: pathlib.PurePath) -> raw_nodes.URI:
        return self._transform_Path(leaf)

    def transform_PurePosixPath(self, leaf: pathlib.PurePosixPath) -> raw_nodes.URI:
        return self._transform_Path(leaf)

    def transform_PureWindowsPath(self, leaf: pathlib.PureWindowsPath) -> raw_nodes.URI:
        return self._transform_Path(leaf)

    def transform_PosixPath(self, leaf: pathlib.PosixPath) -> raw_nodes.URI:
        return self._transform_Path(leaf)

    def transform_WindowsPath(self, leaf: pathlib.WindowsPath) -> raw_nodes.URI:
        return self._transform_Path(leaf)


class RawNodePackageTransformer(NodeTransformer):
    """Transforms raw node fields specified by <node>._include_in_package to local relative paths.
    Adds remote resources to given dictionary."""

    def __init__(
        self, remote_resources: typing.Dict[str, typing.Union[pathlib.PurePath, raw_nodes.URI]], root: pathlib.Path
    ):
        super().__init__()
        self.remote_resources = remote_resources
        self.root = root

    def _transform_resource(
        self, resource: typing.Union[list, pathlib.PurePath, raw_nodes.URI]
    ) -> typing.Union[typing.List[pathlib.Path], _Missing, pathlib.Path]:
        if isinstance(resource, list):
            return [self._transform_resource(r) for r in resource]  # type: ignore  # todo: improve annotation
        elif resource is missing:
            return resource
        elif isinstance(resource, pathlib.PurePath):
            name_from = resource
            if resource.is_absolute():
                folder_in_package = ""
            else:
                if resource.parent.as_posix() == ".":
                    folder_in_package = ""
                else:
                    folder_in_package = resource.parent.as_posix() + "/"

                resource = self.root / resource

        elif isinstance(resource, raw_nodes.URI):
            name_from = pathlib.PurePath(resource.path or "unknown")
            folder_in_package = ""
        else:
            raise TypeError(f"Unexpected type {type(resource)} for {resource}")

        stem = name_from.stem
        suffix = name_from.suffix

        conflict_free_name = f"{folder_in_package}{stem}{suffix}"
        for i in range(100000):
            existing_resource = self.remote_resources.get(conflict_free_name)
            if existing_resource is not None and existing_resource != resource:
                conflict_free_name = f"{folder_in_package}{stem}-{i}{suffix}"
            else:
                break
        else:
            raise ValueError(f"Attempting to pack too many resources with name {stem}{suffix}")

        self.remote_resources[conflict_free_name] = resource

        return pathlib.Path(conflict_free_name)

    def generic_transformer(self, node: GenericRawNode, **kwargs) -> GenericRawNode:
        if isinstance(node, raw_nodes.RawNode):
            resolved_data = {
                field.name: self.transform(getattr(node, field.name), **kwargs) for field in dataclasses.fields(node)
            }
            for incl_field in node._include_in_package:
                field_value = resolved_data[incl_field]
                if field_value is not missing:  # optional fields might be missing
                    resolved_data[incl_field] = self._transform_resource(field_value)

            return dataclasses.replace(node, **resolved_data)
        else:
            return super().generic_transformer(node, **kwargs)


class AbsoluteToRelativePathTransformer(NodeTransformer):
    def __init__(self, *, root_path: os.PathLike):
        self.root_path = pathlib.Path(root_path).resolve()

    def transform_ImportableSourceFile(
        self, node: raw_nodes.ImportableSourceFile, **kwargs
    ) -> raw_nodes.ImportableSourceFile:
        assert isinstance(node.source_file, pathlib.Path)
        sf = node.source_file.relative_to(self.root_path) if node.source_file.is_absolute() else node.source_file
        return raw_nodes.ImportableSourceFile(source_file=sf, callable_name=node.callable_name)

    def _transform_Path(self, leaf: pathlib.Path):
        if leaf.is_absolute():
            return leaf.relative_to(self.root_path)
        else:
            return leaf

    def transform_PosixPath(self, leaf: pathlib.PosixPath, **kwargs) -> pathlib.Path:
        return self._transform_Path(leaf)

    def transform_WindowsPath(self, leaf: pathlib.WindowsPath, **kwargs) -> pathlib.Path:
        return self._transform_Path(leaf)


class RelativeToAbsolutePathTransformer(NodeTransformer):
    def __init__(self, *, root_path: os.PathLike):
        self.root_path = pathlib.Path(root_path).resolve()

    def transform_ImportableSourceFile(
        self, node: raw_nodes.ImportableSourceFile, **kwargs
    ) -> raw_nodes.ImportableSourceFile:
        return raw_nodes.ImportableSourceFile(
            source_file=resolve_local_source(node.source_file, self.root_path), callable_name=node.callable_name
        )

    def _transform_Path(self, leaf: pathlib.Path):
        return self.root_path / leaf

    def transform_PosixPath(self, leaf: pathlib.PosixPath, **kwargs) -> pathlib.Path:
        return self._transform_Path(leaf)

    def transform_WindowsPath(self, leaf: pathlib.WindowsPath, **kwargs) -> pathlib.Path:
        return self._transform_Path(leaf)


class UriNodeTransformer(NodeTransformerKnownParent, RelativeToAbsolutePathTransformer):
    def __init__(self, *, root_path: os.PathLike, uri_only_if_in_package: bool = False):
        super().__init__(root_path=root_path)
        self.uri_only_if_in_package = uri_only_if_in_package

    def transform_URI(
        self, node: raw_nodes.URI, name: typing.Optional[str] = None, parent: typing.Optional[raw_nodes.RawNode] = None
    ) -> typing.Union[raw_nodes.URI, pathlib.Path]:
        if self.uri_only_if_in_package and ((name is None or parent is None) or name not in parent._include_in_package):
            return node
        else:
            local_path = _resolve_source(node, root_path=self.root_path)
            return local_path

    def transform_ImportableSourceFile(
        self, node: raw_nodes.ImportableSourceFile, **kwargs
    ) -> raw_nodes.ResolvedImportableSourceFile:
        return raw_nodes.ResolvedImportableSourceFile(
            source_file=_resolve_source(node.source_file, self.root_path), callable_name=node.callable_name
        )

    def transform_ImportableModule(self, node: raw_nodes.ImportableModule, **kwargs) -> raw_nodes.LocalImportableModule:
        return raw_nodes.LocalImportableModule(**dataclasses.asdict(node), root_path=self.root_path)



import dataclasses
from pathlib import Path
from typing import List, Sequence, Tuple, Type

import bioimageio.spec.model
from bioimageio.spec.model.v0_3.schema import _WeightsEntryBase
from bioimageio.spec.shared.utils import resolve_bioimageio_descrcription

try:
    from typing import get_args
except ImportError:
    from typing_extensions import get_args  # type: ignore


@dataclasses.dataclass
class Kwarg:
    name: str
    description: str
    optional: bool
    maybe_optional: bool


@dataclasses.dataclass
class WeightsFormatDocNode:
    name: str
    description: str
    kwargs: List[Kwarg]


def get_doc(schema) -> Tuple[List[Kwarg], List[WeightsFormatDocNode]]:
    """retrieve documentation of weight formats from their definitions schema"""

    def get_kwargs_doc(we: Type[_WeightsEntryBase], exclude: Sequence[str] = tuple()) -> List[Kwarg]:
        return sorted(
            [
                Kwarg(
                    name=name,
                    description=resolve_bioimageio_descrcription(f.bioimageio_description),
                    optional=not f.required or bool(f.missing),
                    maybe_optional=f.bioimageio_maybe_required,
                )
                for name, f in we().fields.items()
                if name != "weights_format" and name not in exclude
            ],
            key=lambda kw: (kw.optional, kw.name),
        )

    common_kwargs = get_kwargs_doc(_WeightsEntryBase)

    def get_wf_name_from_wf_schema(wfs):
        return wfs().fields["weights_format"].validate.comparable

    return (
        common_kwargs,
        [
            WeightsFormatDocNode(
                name=get_wf_name_from_wf_schema(wfs),
                description=resolve_bioimageio_descrcription(wfs.bioimageio_description),
                kwargs=get_kwargs_doc(wfs, exclude=[kw.name for kw in common_kwargs]),
            )
            for wfs in get_args(schema.WeightsEntry)  # schema.WeightsEntry is a typing.Union of weights format schemas
        ],
    )


def get_md_kwargs(kwargs: Sequence[Kwarg], indent: int = 0):
    md = ""
    for kwarg in kwargs:
        md += (
            f"{' ' * indent}- `{kwarg.name}` _{'optional' if kwarg.optional or kwarg.maybe_optional else 'required'}"
            f"{'*' if kwarg.maybe_optional else ''}_ {kwarg.description}\n"
        )

    return md


def md_from_doc(doc_nodes: List[WeightsFormatDocNode]):
    md = ""
    for doc in doc_nodes:
        md += f"### `{doc.name}`\n{doc.description}\n"
        if doc.kwargs:
            md += f"- key word arguments:\n"
            md += get_md_kwargs(doc.kwargs, indent=2)

    return md


def export_markdown_docs(folder: Path, spec) -> None:
    model_or_version = spec.__name__.split(".")[-1]
    format_version_wo_patch = ".".join(spec.format_version.split(".")[:2])
    if model_or_version == "model":
        format_version_file_name = "latest"
    else:
        format_version_file_name = format_version_wo_patch.replace(".", "_")

    common_kwargs, doc = get_doc(spec.schema)
    md = (
        (
            f"# Weight formats in model spec {format_version_wo_patch}\n"
            "## Common key word arguments for all weight formats\n"
            "Optional arguments are marked as _optional_ or _optional*_ with an asterisk if they are optional "
            "depending on another argument's value.\n\n"
        )
        + get_md_kwargs(common_kwargs)
        + ("\n## Weight formats and their additional key word arguments\n")
    )
    md += md_from_doc(doc)
    path = folder / f"weight_formats_spec_{format_version_file_name}.md"
    path.write_text(md, encoding="utf-8")


if __name__ == "__main__":
    import bioimageio.spec.model

    dist = Path(__file__).parent / "../dist"
    dist.mkdir(exist_ok=True)

    export_markdown_docs(dist, bioimageio.spec.model.v0_3)
    export_markdown_docs(dist, bioimageio.spec.model.v0_4)
    export_markdown_docs(dist, bioimageio.spec.model)



from . import collection, model, rdf, shared
from .commands import validate
from .io_ import (
    get_resource_package_content,
    load_raw_resource_description,
    serialize_raw_resource_description,
    serialize_raw_resource_description_to_dict,
)
from .v import __version__



import packaging.version
from dataclasses import dataclass
from datetime import datetime
from pathlib import Path
from typing import Any, Dict, List, Union

from marshmallow import missing
from marshmallow.utils import _Missing

from bioimageio.spec.dataset.v0_2.raw_nodes import Dataset
from bioimageio.spec.model.v0_3.raw_nodes import (
    InputTensor,
    KerasHdf5WeightsEntry as KerasHdf5WeightsEntry03,
    OnnxWeightsEntry as OnnxWeightsEntry03,
    OutputTensor,
    Postprocessing,
    PostprocessingName,
    Preprocessing,
    PreprocessingName,
    RunMode,
    TensorflowJsWeightsEntry as TensorflowJsWeightsEntry03,
    TensorflowSavedModelBundleWeightsEntry as TensorflowSavedModelBundleWeightsEntry03,
    _WeightsEntryBase as _WeightsEntryBase03,
)
from bioimageio.spec.rdf.v0_2.raw_nodes import Author, Maintainer, RDF as _RDF
from bioimageio.spec.shared.raw_nodes import (
    Dependencies,
    ImplicitOutputShape,
    ImportableModule,
    ImportableSourceFile,
    ParametrizedInputShape,
    RawNode,
    URI,
)

try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal  # type: ignore

# reassign to use imported classes
ImplicitOutputShape = ImplicitOutputShape
Maintainer = Maintainer
ParametrizedInputShape = ParametrizedInputShape
Postprocessing = Postprocessing
PostprocessingName = PostprocessingName
Preprocessing = Preprocessing
PreprocessingName = PreprocessingName

FormatVersion = Literal[
    "0.4.0", "0.4.1", "0.4.2", "0.4.3", "0.4.4", "0.4.5"
]  # newest format needs to be last (used in __init__.py)
WeightsFormat = Literal[
    "pytorch_state_dict", "torchscript", "keras_hdf5", "tensorflow_js", "tensorflow_saved_model_bundle", "onnx"
]

ImportableSource = Union[ImportableSourceFile, ImportableModule]


@dataclass
class _WeightsEntryBase(_WeightsEntryBase03):
    dependencies: Union[_Missing, Dependencies] = missing


@dataclass
class KerasHdf5WeightsEntry(_WeightsEntryBase, KerasHdf5WeightsEntry03):
    pass


@dataclass
class OnnxWeightsEntry(_WeightsEntryBase, OnnxWeightsEntry03):
    pass


@dataclass
class PytorchStateDictWeightsEntry(_WeightsEntryBase):
    weights_format_name = "Pytorch State Dict"
    architecture: ImportableSource = missing
    architecture_sha256: Union[_Missing, str] = missing
    kwargs: Union[_Missing, Dict[str, Any]] = missing
    pytorch_version: Union[_Missing, packaging.version.Version] = missing


@dataclass
class TensorflowJsWeightsEntry(_WeightsEntryBase, TensorflowJsWeightsEntry03):
    pass


@dataclass
class TensorflowSavedModelBundleWeightsEntry(_WeightsEntryBase, TensorflowSavedModelBundleWeightsEntry03):
    pass


@dataclass
class TorchscriptWeightsEntry(_WeightsEntryBase):
    weights_format_name = "Torchscript"
    pytorch_version: Union[_Missing, packaging.version.Version] = missing


WeightsEntry = Union[
    KerasHdf5WeightsEntry,
    OnnxWeightsEntry,
    PytorchStateDictWeightsEntry,
    TensorflowJsWeightsEntry,
    TensorflowSavedModelBundleWeightsEntry,
    TorchscriptWeightsEntry,
]


@dataclass
class LinkedDataset(RawNode):
    id: str


@dataclass
class ModelParent(RawNode):
    id: Union[_Missing, str] = missing
    uri: Union[_Missing, URI, Path] = missing
    sha256: Union[_Missing, str] = missing


@dataclass
class Model(_RDF):
    _include_in_package = ("covers", "documentation", "test_inputs", "test_outputs", "sample_inputs", "sample_outputs")

    format_version: FormatVersion = missing
    inputs: List[InputTensor] = missing
    license: str = missing
    outputs: List[OutputTensor] = missing
    packaged_by: Union[_Missing, List[Author]] = missing
    parent: Union[_Missing, ModelParent] = missing
    run_mode: Union[_Missing, RunMode] = missing
    sample_inputs: Union[_Missing, List[Union[URI, Path]]] = missing
    sample_outputs: Union[_Missing, List[Union[URI, Path]]] = missing
    test_inputs: List[Union[URI, Path]] = missing
    test_outputs: List[Union[URI, Path]] = missing
    timestamp: datetime = missing
    training_data: Union[_Missing, Dataset, LinkedDataset] = missing
    type: Literal["model"] = missing
    weights: Dict[WeightsFormat, WeightsEntry] = missing



""" raw nodes for the general RDF spec

raw nodes are the deserialized equivalent to the content of any RDF.
serialization and deserialization are defined in schema:
RDF <--schema--> raw nodes
"""
import dataclasses
import pathlib

import packaging.version
import warnings
from dataclasses import dataclass
from pathlib import Path
from typing import Any, Dict, List, Union

from marshmallow import missing
from marshmallow.utils import _Missing

from bioimageio.spec.shared.raw_nodes import RawNode, ResourceDescription, URI

try:
    from typing import Literal, get_args
except ImportError:
    from typing_extensions import Literal, get_args  # type: ignore

FormatVersion = Literal[
    "0.2.0", "0.2.1", "0.2.2"
]  # newest format needs to be last (used to determine latest format version)


@dataclass(init=False)
class Attachments(RawNode):
    _include_in_package = ("files",)

    files: Union[_Missing, List[Union[Path, URI]]] = missing
    unknown: Dict[str, Any] = missing

    def __init__(
        self,
        files: Union[_Missing, List[Union[Path, URI]]] = missing,
        unknown: Dict[str, Any] = missing,
        **implicitly_unknown,
    ):
        self.files = files
        self.unknown = unknown or {}
        self.unknown.update(implicitly_unknown)
        super().__init__()


@dataclass
class _Person(RawNode):
    name: Union[_Missing, str] = missing
    affiliation: Union[_Missing, str] = missing
    email: Union[_Missing, str] = missing
    github_user: Union[_Missing, str] = missing
    orcid: Union[_Missing, str] = missing


@dataclass
class Author(_Person):
    name: str = missing


@dataclass
class Maintainer(_Person):
    github_user: str = missing


@dataclass
class CiteEntry(RawNode):
    text: str = missing
    doi: Union[_Missing, str] = missing
    url: Union[_Missing, str] = missing


@dataclass
class Badge(RawNode):
    label: str = missing
    icon: Union[_Missing, str] = missing
    url: Union[_Missing, URI, Path] = missing


@dataclass
class RDF(ResourceDescription):
    attachments: Union[_Missing, Attachments] = missing
    authors: Union[_Missing, List[Author]] = missing
    badges: Union[_Missing, List[Badge]] = missing
    cite: Union[_Missing, List[CiteEntry]] = missing
    config: Union[_Missing, dict] = missing
    covers: Union[_Missing, List[Union[URI, Path]]] = missing
    description: str = missing
    documentation: Union[_Missing, Path, URI] = missing
    download_url: Union[_Missing, URI] = missing
    format_version: FormatVersion = missing
    git_repo: Union[_Missing, str] = missing
    id: Union[_Missing, str] = missing
    icon: Union[_Missing, str] = missing
    license: Union[_Missing, str] = missing
    links: Union[_Missing, List[str]] = missing
    maintainers: Union[_Missing, List[Maintainer]] = missing
    rdf_source: Union[_Missing, URI] = missing
    source: Union[_Missing, URI, Path] = missing
    tags: Union[_Missing, List[str]] = missing

    # manual __init__ to allow for unknown kwargs
    def __init__(
        self,
        *,
        # ResourceDescription
        format_version: FormatVersion,
        name: str,
        type: str = missing,
        version: Union[_Missing, packaging.version.Version] = missing,
        root_path: pathlib.Path = pathlib.Path(),
        # RDF
        attachments: Union[_Missing, Dict[str, Any]] = missing,
        authors: Union[_Missing, List[Author]] = missing,
        badges: Union[_Missing, List[Badge]] = missing,
        cite: Union[_Missing, List[CiteEntry]] = missing,
        config: Union[_Missing, dict] = missing,
        covers: Union[_Missing, List[Union[URI, Path]]] = missing,
        description: str,
        documentation: Union[_Missing, Path, URI] = missing,
        download_url: Union[_Missing, URI] = missing,
        git_repo: Union[_Missing, str] = missing,
        id: Union[_Missing, str] = missing,
        icon: Union[_Missing, str] = missing,
        license: Union[_Missing, str] = missing,
        links: Union[_Missing, List[str]] = missing,
        maintainers: Union[_Missing, List[Maintainer]] = missing,
        rdf_source: Union[_Missing, URI] = missing,
        source: Union[_Missing, URI, Path] = missing,
        tags: Union[_Missing, List[str]] = missing,
        **unknown_kwargs,
    ):
        self.attachments = attachments
        self.authors = authors
        self.badges = badges
        self.cite = cite
        self.config = config
        self.covers = covers
        self.description = description
        self.documentation = documentation
        self.download_url = download_url
        self.git_repo = git_repo
        self.id = id
        self.icon = icon
        self.license = license
        self.links = links
        self.maintainers = maintainers
        self.rdf_source = rdf_source
        self.source = source
        self.tags = tags
        super().__init__(format_version=format_version, name=name, type=type, version=version, root_path=root_path)

        if unknown_kwargs:
            # make sure we didn't forget a defined field
            field_names = set(f.name for f in dataclasses.fields(self))
            for uk in unknown_kwargs:
                assert uk not in field_names, uk

            warnings.warn(f"discarding unknown RDF fields: {unknown_kwargs}")

    def __post_init__(self):
        if self.type is missing:
            self.type = self.__class__.__name__.lower()

        super().__post_init__()



"""simple io functionality to deserialize a resource description from a resource description file (RDF)
(in form of a dict, e.g. from yaml.load('rdf.yaml') to a raw_nodes.ResourceDescription raw node,
which is a python dataclass
"""
import os
import pathlib
import warnings
import zipfile
from hashlib import sha256
from io import StringIO
from tempfile import TemporaryDirectory
from types import ModuleType
from typing import Dict, IO, Optional, Sequence, Tuple, Union

from marshmallow import missing

from bioimageio.spec.shared import RDF_NAMES, raw_nodes, resolve_rdf_source, resolve_rdf_source_and_type, resolve_source
from bioimageio.spec.shared.common import (
    BIOIMAGEIO_CACHE_PATH,
    BIOIMAGEIO_USE_CACHE,
    get_class_name_from_type,
    get_format_version_module,
    get_latest_format_version,
    get_latest_format_version_module,
    no_cache_tmp_list,
    yaml,
)
from bioimageio.spec.shared.node_transformer import (
    AbsoluteToRelativePathTransformer,
    GenericRawNode,
    GenericRawRD,
    PathToRemoteUriTransformer,
    RawNodePackageTransformer,
    RelativeToAbsolutePathTransformer,
)
from bioimageio.spec.shared.raw_nodes import ResourceDescription as RawResourceDescription
from bioimageio.spec.shared.schema import SharedBioImageIOSchema

try:
    from typing import Protocol
except ImportError:
    from typing_extensions import Protocol  # type: ignore


LATEST = "latest"


class ConvertersModule(Protocol):
    def maybe_convert(self, data: dict) -> dict:
        raise NotImplementedError


class SubModuleUtils(Protocol):
    def filter_resource_description(self, raw_rd: GenericRawNode, **kwargs) -> GenericRawNode:
        raise NotImplementedError


class SpecSubmodule(Protocol):
    format_version: str

    converters: ConvertersModule
    nodes: ModuleType
    raw_nodes: ModuleType
    schema: ModuleType
    utils: SubModuleUtils


def _get_spec_submodule(type_: str, data_version: str = LATEST) -> SpecSubmodule:
    if not isinstance(data_version, str):
        raise TypeError(f"invalid 'format_version' {data_version}")

    if data_version == LATEST:
        sub_spec = get_latest_format_version_module(type_)
    else:
        sub_spec = get_format_version_module(type_, data_version)

    return sub_spec


def extract_resource_package(
    source: Union[os.PathLike, IO, str, bytes, raw_nodes.URI]
) -> Tuple[dict, str, pathlib.Path]:
    """extract a zip source to BIOIMAGEIO_CACHE_PATH"""
    source, source_name, root = resolve_rdf_source(source)
    if isinstance(root, bytes):
        raise NotImplementedError("package source was bytes")

    if BIOIMAGEIO_USE_CACHE:
        package_path = BIOIMAGEIO_CACHE_PATH / "extracted_packages" / sha256(str(root).encode("utf-8")).hexdigest()
        package_path.mkdir(exist_ok=True, parents=True)
    else:
        tmp_dir = TemporaryDirectory()
        no_cache_tmp_list.append(tmp_dir)
        package_path = pathlib.Path(tmp_dir.name)

    if isinstance(root, raw_nodes.URI):
        for rdf_name in RDF_NAMES:
            if (package_path / rdf_name).exists():
                download = None
                break
        else:
            download = resolve_source(root)

        local_source = download
    else:
        download = None
        local_source = root

    if local_source is not None:
        with zipfile.ZipFile(local_source) as zf:
            zf.extractall(package_path)

    for rdf_name in RDF_NAMES:
        if (package_path / rdf_name).exists():
            break
    else:
        raise FileNotFoundError(f"Missing 'rdf.yaml' in {root} extracted from {download}")

    if download is not None:
        try:
            os.remove(download)
        except Exception as e:
            warnings.warn(f"Could not remove download {download} due to {e}")

    assert isinstance(package_path, pathlib.Path)
    return source, source_name, package_path


def _replace_relative_paths_for_remote_source(raw_rd: RawResourceDescription) -> RawResourceDescription:
    root = raw_rd.root_path
    if isinstance(root, raw_nodes.URI):
        # for a remote source relative paths are invalid; replace all relative file paths in source with URLs
        raw_rd = PathToRemoteUriTransformer(remote_source=root).transform(raw_rd)
        root_path = pathlib.Path()  # root_path cannot be URI
    elif isinstance(root, pathlib.Path):
        if zipfile.is_zipfile(root):
            _, _, root_path = extract_resource_package(root)
        else:
            root_path = root
    elif isinstance(root, bytes):
        root_path = pathlib.Path()
    else:
        raise TypeError(root)

    assert isinstance(root_path, pathlib.Path)
    raw_rd.root_path = root_path.resolve()
    return raw_rd


def load_raw_resource_description(
    source: Union[dict, os.PathLike, IO, str, bytes, raw_nodes.URI, RawResourceDescription],
    update_to_format: Optional[str] = None,
) -> RawResourceDescription:
    """load a raw python representation from a BioImage.IO resource description.
    Use `bioimageio.core.load_resource_description` for a more convenient representation of the resource.
    and `bioimageio.core.load_raw_resource_description` to ensure the 'root_path' attribute of the returned object is
    a local file path.

    Args:
        source: resource description or resource description file (RDF)
        update_to_format: update resource to specific major.minor format version; ignoring patch version.
    Returns:
        raw BioImage.IO resource
    """
    root = None
    if isinstance(source, RawResourceDescription):
        if update_to_format == "latest":
            update_to_format = get_latest_format_version(source.type)

        if update_to_format is not None and source.format_version != update_to_format:
            # do serialization round-trip to account for 'update_to_format' but keep root_path
            root = source.root_path
            source = serialize_raw_resource_description_to_dict(source)
        else:
            return source

    data, source_name, _root, type_ = resolve_rdf_source_and_type(source)
    if root is None:
        root = _root

    class_name = get_class_name_from_type(type_)

    if update_to_format is None:
        data_version = data.get("format_version", LATEST)
    elif update_to_format == LATEST:
        data_version = LATEST
    else:
        data_version = ".".join(update_to_format.split("."[:2]))
        if update_to_format.count(".") > 1:
            warnings.warn(
                f"Ignoring patch version of update_to_format {update_to_format} "
                f"(always updating to latest patch version)."
            )

    sub_spec = _get_spec_submodule(type_, data_version)
    schema: SharedBioImageIOSchema = getattr(sub_spec.schema, class_name)()

    data = sub_spec.converters.maybe_convert(data)
    raw_rd = schema.load(data)
    raw_rd.root_path = root

    raw_rd = _replace_relative_paths_for_remote_source(raw_rd)
    raw_rd = RelativeToAbsolutePathTransformer(root_path=raw_rd.root_path).transform(raw_rd)

    return raw_rd


def serialize_raw_resource_description_to_dict(
    raw_rd: RawResourceDescription, convert_absolute_paths: bool = False
) -> dict:
    """serialize a raw nodes resource description to a dict with the content of a resource description file (RDF).
    If 'convert_absolute_paths' all absolute paths are converted to paths relative to raw_rd.root_path before
    serialization.
    """
    class_name = get_class_name_from_type(raw_rd.type)
    sub_spec = _get_spec_submodule(raw_rd.type, raw_rd.format_version)
    schema: SharedBioImageIOSchema = getattr(sub_spec.schema, class_name)()

    if convert_absolute_paths:
        raw_rd = AbsoluteToRelativePathTransformer(root_path=raw_rd.root_path).transform(raw_rd)

    serialized = schema.dump(raw_rd)
    assert isinstance(serialized, dict)
    assert missing not in serialized.values()

    return serialized


def serialize_raw_resource_description(raw_rd: RawResourceDescription, convert_absolute_paths: bool = True) -> str:
    if yaml is None:
        raise RuntimeError("'serialize_raw_resource_description' requires yaml")

    serialized = serialize_raw_resource_description_to_dict(raw_rd, convert_absolute_paths=convert_absolute_paths)

    with StringIO() as stream:
        yaml.dump(serialized, stream)
        return stream.getvalue()


def save_raw_resource_description(raw_rd: RawResourceDescription, path: pathlib.Path):
    if yaml is None:
        raise RuntimeError("'save_raw_resource_description' requires yaml")

    warnings.warn("only saving serialized rdf, no associated resources.")
    if path.suffix != ".yaml":
        warnings.warn("saving with '.yaml' suffix is strongly encouraged.")

    serialized = serialize_raw_resource_description_to_dict(raw_rd)
    yaml.dump(serialized, path)


def get_resource_package_content_wo_rdf(
    raw_rd: Union[GenericRawRD, raw_nodes.URI, str, pathlib.Path],
    *,
    weights_priority_order: Optional[Sequence[str]] = None,  # model only
) -> Tuple[GenericRawNode, Dict[str, Union[pathlib.PurePath, raw_nodes.URI]]]:
    """
    Args:
        raw_rd: raw resource description
        # for model resources only:
        weights_priority_order: If given only the first weights format present in the model is included.
                                If none of the prioritized weights formats is found all are included.

    Returns:
        Tuple of updated raw resource description and package content of remote URIs, local file paths or text content
        keyed by file names.
        Important note: the serialized rdf.yaml is not included.
    """
    if not isinstance(raw_rd, raw_nodes.ResourceDescription):
        raw_rd = load_raw_resource_description(raw_rd)

    sub_spec = _get_spec_submodule(raw_rd.type, raw_rd.format_version)
    if raw_rd.type == "model":
        filter_kwargs = dict(weights_priority_order=weights_priority_order)
    else:
        filter_kwargs = {}

    raw_rd = sub_spec.utils.filter_resource_description(raw_rd, **filter_kwargs)

    content: Dict[str, Union[pathlib.PurePath, raw_nodes.URI, str]] = {}
    raw_rd = RawNodePackageTransformer(content, raw_rd.root_path).transform(raw_rd)
    assert "rdf.yaml" not in content
    return raw_rd, content


def get_resource_package_content(
    raw_rd: Union[GenericRawNode, raw_nodes.URI, str, pathlib.Path],
    *,
    weights_priority_order: Optional[Sequence[str]] = None,  # model only
) -> Dict[str, Union[str, pathlib.PurePath, raw_nodes.URI]]:
    """
    Args:
        raw_rd: raw resource description
        # for model resources only:
        weights_priority_order: If given only the first weights format present in the model is included.
                                If none of the prioritized weights formats is found all are included.

    Returns:
        Package content of remote URIs, local file paths or text content keyed by file names.
    """
    if yaml is None:
        raise RuntimeError(
            "'get_resource_package_content' requires yaml; note that 'get_resource_package_content_wo_rdf' may be used "
            "without yaml"
        )

    content: Dict[str, Union[str, pathlib.PurePath, raw_nodes.URI]]
    raw_rd, content = get_resource_package_content_wo_rdf(raw_rd, weights_priority_order=weights_priority_order)
    content["rdf.yaml"] = serialize_raw_resource_description(raw_rd)
    return content



def is_valid_orcid_id(orcid_id: str):
    """adapted from stdnum.iso7064.mod_11_2.checksum()"""
    check = 0
    for n in orcid_id:
        check = (2 * check + int(10 if n == "X" else n)) % 11
    return check == 1



from marshmallow import missing, validates

from bioimageio.spec.rdf.v0_2.schema import RDF
from bioimageio.spec.shared import fields
from bioimageio.spec.shared.schema import SharedBioImageIOSchema, WithUnknown
from . import raw_nodes

try:
    from typing import List, Union, get_args
except ImportError:
    from typing_extensions import get_args  # type: ignore


class _BioImageIOSchema(SharedBioImageIOSchema):
    raw_nodes = raw_nodes


class CollectionEntry(_BioImageIOSchema, WithUnknown):
    field_name_unknown_dict = "rdf_update"
    rdf_source = fields.Union([fields.URL(), fields.DOI()])


class Collection(_BioImageIOSchema, WithUnknown, RDF):
    bioimageio_description = f"""# BioImage.IO Collection Resource Description File Specification {get_args(raw_nodes.FormatVersion)[-1]}
This specification defines the fields used in a BioImage.IO-compliant resource description file (`RDF`) for describing collections of other resources.
These fields are typically stored in a YAML file which we call Collection Resource Description File or `collection RDF`.

The collection RDF YAML file contains mandatory and optional fields. In the following description, optional fields are indicated by _optional_.
_optional*_ with an asterisk indicates the field is optional depending on the value in another field.
"""
    collection = fields.List(
        fields.Nested(CollectionEntry()),
        bioimageio_description="Collection entries. Each entry needs to specify a valid RDF with an id. "
        "Each collection entry RDF is based on the collection RDF itself, "
        "updated by rdf_source content if rdf_source is specified, "
        "and updated by any fields specified directly in the entry. "
        "In this context 'update' refers to overwriting RDF root fields by name."
        "Except for the `id` field, which appends to the collection RDF `id` "
        "such that full_collection_entry_id=<collection_id>/<entry_id>",
        required=True,
    )

    @validates("collection")
    def unique_ids(self, value: List[Union[dict, raw_nodes.CollectionEntry]]):
        ids = [
            (v.get("id", missing), v.get("rdf_source", missing))
            if isinstance(v, dict)
            else (v.rdf_update.get("id", missing), v.rdf_source)
            for v in value
        ]
        # skip check for id only specified in remote source
        ids = [vid for vid, vs in ids if not (vid is missing and vs is not missing)]

        if missing in ids:
            raise ValueError(f"Missing ids in collection entries")

        non_string_ids = [v for v in ids if not isinstance(v, str)]
        if non_string_ids:
            raise ValueError(f"Non-string ids in collection: {non_string_ids}")

        seen = set()
        duplicates = []
        for v in ids:
            if v in seen:
                duplicates.append(v)
            else:
                seen.add(v)

        if duplicates:
            raise ValueError(f"Duplicate ids in collection: {duplicates}")



import json
import os
import pathlib
import re
import shutil
import typing
import warnings
import zipfile
from functools import singledispatch
from io import BytesIO, StringIO
from tempfile import TemporaryDirectory
from urllib.request import url2pathname, urlopen

from marshmallow import ValidationError

from . import fields, raw_nodes
from .common import (
    BIOIMAGEIO_CACHE_PATH,
    BIOIMAGEIO_COLLECTION_URL,
    BIOIMAGEIO_USE_CACHE,
    BIOIMAGEIO_SITE_CONFIG_URL,
    DOI_REGEX,
    get_spec_type_from_type,
    no_cache_tmp_list,
    RDF_NAMES,
    tqdm,
    yaml,
)
from .raw_nodes import URI


def _is_path(s: typing.Any) -> bool:
    if not isinstance(s, (str, os.PathLike)):
        return False

    try:
        return pathlib.Path(s).exists()
    except OSError:
        return False


class RDF_Source(typing.NamedTuple):
    data: dict
    name: str
    root: typing.Union[pathlib.Path, raw_nodes.URI, bytes]


def resolve_rdf_source(
    source: typing.Union[dict, os.PathLike, typing.IO, str, bytes, URI, raw_nodes.ResourceDescription]
) -> RDF_Source:
    # reduce possible source types
    if isinstance(source, (BytesIO, StringIO)):
        source = source.getvalue()
    elif isinstance(source, os.PathLike):
        source = pathlib.Path(source)
    elif isinstance(source, raw_nodes.URI):
        source = str(source)
    elif isinstance(source, raw_nodes.ResourceDescription):
        from bioimageio.spec.io_ import serialize_raw_resource_description_to_dict

        source = serialize_raw_resource_description_to_dict(source)

    if not isinstance(source, (dict, pathlib.Path, str, bytes)):
        raise TypeError(f"Unexpected source type {type(source)}")

    if isinstance(source, pathlib.Path):
        source_name = str(source)
        root: typing.Union[pathlib.Path, raw_nodes.URI, bytes] = source.parent
    elif isinstance(source, dict):
        source_name = f"{{name: {source.get('name', '<unknown>')}, ...}}"
        source = dict(source)
        given_root = source.pop("root_path", pathlib.Path())
        if _is_path(given_root):
            root = pathlib.Path(given_root)
        elif isinstance(given_root, URI):
            root = given_root
        elif isinstance(given_root, str):
            root = URI(uri_string=given_root)
        else:
            raise ValueError(f"Encountered invalid root {given_root}")
    elif isinstance(source, (str, bytes)):
        source_name = str(source[:120]) + "..."
        # string might be path or yaml string; for yaml string (or bytes) set root to cwd

        if _is_path(source):
            assert isinstance(source, (str, os.PathLike))
            root = pathlib.Path(source).parent
        else:
            root = pathlib.Path()
    else:
        raise TypeError(source)

    if isinstance(source, str):
        # source might be bioimageio id, doi, url or file path -> resolve to pathlib.Path

        bioimageio_rdf_source: typing.Optional[str] = (BIOIMAGEIO_COLLECTION_ENTRIES or {}).get(source, (None, None))[1]

        if bioimageio_rdf_source is not None:
            # source is bioimageio id
            source = bioimageio_rdf_source
        elif re.fullmatch(DOI_REGEX, source):  # turn doi into url
            import requests  # not available in pyodide

            zenodo_prefix = "10.5281/zenodo."
            zenodo_record_api = "https://zenodo.org/api/records"
            zenodo_sandbox_prefix = "10.5072/zenodo."
            zenodo_sandbox_record_api = "https://sandbox.zenodo.org/api/records"
            is_zenodo_doi = False
            if source.startswith(zenodo_prefix):
                is_zenodo_doi = True
            elif source.startswith(zenodo_sandbox_prefix):
                # zenodo sandbox doi (which is not a valid doi)
                zenodo_prefix = zenodo_sandbox_prefix
                zenodo_record_api = zenodo_sandbox_record_api
                is_zenodo_doi = True

            if is_zenodo_doi:
                # source is a doi pointing to a zenodo record;
                # we'll expect an rdf.yaml file in that record and use it as source...
                record_id = source[len(zenodo_prefix) :]
                s_count = record_id.count("/")
                if s_count:
                    # record_id/record_version_id
                    if s_count != 1:
                        warnings.warn(
                            f"Unexpected Zenodo record ids: {record_id}. "
                            f"Expected <concept id> or <concept id>/<version id>."
                        )

                    record_id = record_id.split("/")[-1]

                response = requests.get(f"{zenodo_record_api}/{record_id}")
                if not response.ok:
                    raise RuntimeError(response.status_code)

                zenodo_record = response.json()
                for rdf_name in RDF_NAMES:
                    for f in zenodo_record["files"]:
                        if f["key"] == rdf_name:
                            source = f["links"]["self"]
                            break
                    else:
                        continue

                    break
                else:
                    raise ValidationError(f"No RDF found; looked for {RDF_NAMES}")

            else:
                # resolve doi
                # todo: make sure the resolved url points to a rdf.yaml or a zipped package
                response = urlopen(f"https://doi.org/{source}?type=URL")
                source = response.url
                assert isinstance(source, str)
                if not (source.endswith(".yaml") or source.endswith(".zip")):
                    raise NotImplementedError(
                        f"Resolved doi {source_name} to {source}, but don't know where to find 'rdf.yaml' "
                        f"or a packaged resource zip file."
                    )

        assert isinstance(source, str)
        if source.startswith("http"):
            root = raw_nodes.URI(uri_string=source)
            source = _download_url(root)

        if _is_path(source):
            source = pathlib.Path(source)

    if isinstance(source, (pathlib.Path, str, bytes)):
        # source is either:
        #   - a file path (to a yaml or a packaged zip)
        #   - a yaml string,
        #   - or yaml file or zip package content as bytes

        if yaml is None:
            raise RuntimeError(f"Cannot read RDF from {source_name} without ruamel.yaml dependency!")

        if isinstance(source, bytes):
            potential_package: typing.Union[pathlib.Path, typing.IO, str] = BytesIO(source)
            potential_package.seek(0)  # type: ignore
        else:
            potential_package = source

        if zipfile.is_zipfile(potential_package):
            with zipfile.ZipFile(potential_package) as zf:
                for rdf_name in RDF_NAMES:
                    if rdf_name in zf.namelist():
                        break
                else:
                    raise ValueError(f"Missing 'rdf.yaml' in package {source_name}")

                assert isinstance(source, (pathlib.Path, bytes))
                root = source
                source = BytesIO(zf.read(rdf_name))

        source = yaml.load(source)

    if not isinstance(source, dict):
        raise TypeError(f"Expected dict type for loaded source, but got: {type(source)}")

    return RDF_Source(source, source_name, root)


def resolve_rdf_source_and_type(
    source: typing.Union[os.PathLike, str, dict, raw_nodes.URI]
) -> typing.Tuple[dict, str, typing.Union[pathlib.Path, raw_nodes.URI, bytes], str]:
    data, source_name, root = resolve_rdf_source(source)

    type_ = get_spec_type_from_type(data.get("type"))
    return data, source_name, root, type_


@singledispatch  # todo: fix type annotations
def resolve_source(source, root_path: os.PathLike = pathlib.Path(), output=None):
    raise TypeError(type(source))


@resolve_source.register
def _resolve_source_uri_node(
    source: raw_nodes.URI, root_path: os.PathLike = pathlib.Path(), output: typing.Optional[os.PathLike] = None
) -> pathlib.Path:
    path_or_remote_uri = resolve_local_source(source, root_path, output)
    if isinstance(path_or_remote_uri, raw_nodes.URI):
        local_path = _download_url(path_or_remote_uri, output)
    elif isinstance(path_or_remote_uri, pathlib.Path):
        local_path = path_or_remote_uri
    else:
        raise TypeError(path_or_remote_uri)

    return local_path


@resolve_source.register
def _resolve_source_str(
    source: str, root_path: os.PathLike = pathlib.Path(), output: typing.Optional[os.PathLike] = None
) -> pathlib.Path:
    return resolve_source(fields.Union([fields.URI(), fields.Path()]).deserialize(source), root_path, output)


@resolve_source.register
def _resolve_source_path(
    source: pathlib.Path, root_path: os.PathLike = pathlib.Path(), output: typing.Optional[os.PathLike] = None
) -> pathlib.Path:
    if not source.is_absolute():
        source = pathlib.Path(root_path).absolute() / source

    if output is None:
        return source
    else:
        try:
            shutil.copyfile(source, output)
        except shutil.SameFileError:  # source and output are identical
            pass
        return pathlib.Path(output)


@resolve_source.register
def _resolve_source_resolved_importable_path(
    source: raw_nodes.ResolvedImportableSourceFile,
    root_path: os.PathLike = pathlib.Path(),
    output: typing.Optional[os.PathLike] = None,
) -> raw_nodes.ResolvedImportableSourceFile:
    return raw_nodes.ResolvedImportableSourceFile(
        callable_name=source.callable_name, source_file=resolve_source(source.source_file, root_path, output)
    )


@resolve_source.register
def _resolve_source_importable_path(
    source: raw_nodes.ImportableSourceFile,
    root_path: os.PathLike = pathlib.Path(),
    output: typing.Optional[os.PathLike] = None,
) -> raw_nodes.ResolvedImportableSourceFile:
    return raw_nodes.ResolvedImportableSourceFile(
        callable_name=source.callable_name, source_file=resolve_source(source.source_file, root_path, output)
    )


@resolve_source.register
def _resolve_source_list(
    source: list,
    root_path: os.PathLike = pathlib.Path(),
    output: typing.Optional[typing.Sequence[typing.Optional[os.PathLike]]] = None,
) -> typing.List[pathlib.Path]:
    assert output is None or len(output) == len(source)
    return [resolve_source(el, root_path, out) for el, out in zip(source, output or [None] * len(source))]


def resolve_local_sources(
    sources: typing.Sequence[typing.Union[str, os.PathLike, raw_nodes.URI]],
    root_path: os.PathLike,
    outputs: typing.Optional[typing.Sequence[typing.Optional[os.PathLike]]] = None,
) -> typing.List[typing.Union[pathlib.Path, raw_nodes.URI]]:
    if outputs is None:
        outputs = [None] * len(sources)

    assert outputs is not None
    assert len(outputs) == len(sources)
    return [resolve_local_source(src, root_path, out) for src, out in zip(sources, outputs)]


def resolve_local_source(
    source: typing.Union[str, os.PathLike, raw_nodes.URI],
    root_path: os.PathLike,
    output: typing.Optional[os.PathLike] = None,
) -> typing.Union[pathlib.Path, raw_nodes.URI]:
    if isinstance(source, os.PathLike) or isinstance(source, str):
        try:  # source as path from cwd
            is_path_cwd = pathlib.Path(source).exists()
        except OSError:
            is_path_cwd = False

        try:  # source as relative path from root_path
            path_from_root = pathlib.Path(root_path) / source
            is_path_rp = (path_from_root).exists()
        except OSError:
            is_path_rp = False
        else:
            if not is_path_cwd and is_path_rp:
                source = path_from_root

        if is_path_cwd or is_path_rp:
            source = pathlib.Path(source)
            if output is None:
                return source
            else:
                try:
                    shutil.copyfile(source, output)
                except shutil.SameFileError:
                    pass
                return pathlib.Path(output)

        elif isinstance(source, os.PathLike):
            raise FileNotFoundError(f"Could neither find {source} nor {pathlib.Path(root_path) / source}")

    if isinstance(source, str):
        uri = fields.URI().deserialize(source)
    else:
        uri = source

    assert isinstance(uri, raw_nodes.URI), uri
    if uri.scheme == "file":
        local_path_or_remote_uri: typing.Union[pathlib.Path, raw_nodes.URI] = pathlib.Path(url2pathname(uri.path))
    elif uri.scheme in ("https", "https"):
        local_path_or_remote_uri = uri
    else:
        raise ValueError(f"Unknown uri scheme {uri.scheme}")

    return local_path_or_remote_uri


def source_available(source: typing.Union[pathlib.Path, raw_nodes.URI], root_path: pathlib.Path) -> bool:
    import requests  # not available in pyodide

    local_path_or_remote_uri = resolve_local_source(source, root_path)
    if isinstance(local_path_or_remote_uri, raw_nodes.URI):
        response = requests.head(str(local_path_or_remote_uri))
        available = response.status_code == 200
    elif isinstance(local_path_or_remote_uri, pathlib.Path):
        available = local_path_or_remote_uri.exists()
    else:
        raise TypeError(local_path_or_remote_uri)

    return available


def _download_url(uri: raw_nodes.URI, output: typing.Optional[os.PathLike] = None) -> pathlib.Path:
    if output is not None:
        local_path = pathlib.Path(output)
    elif BIOIMAGEIO_USE_CACHE:
        # todo: proper caching
        local_path = BIOIMAGEIO_CACHE_PATH / uri.scheme / uri.authority / uri.path.strip("/") / uri.query
    else:
        tmp_dir = TemporaryDirectory()
        no_cache_tmp_list.append(tmp_dir)  # keep temporary file until process ends
        local_path = pathlib.Path(tmp_dir.name) / "file"

    if local_path.exists():
        warnings.warn(f"found cached {local_path}. Skipping download of {uri}.")
    else:
        local_path.parent.mkdir(parents=True, exist_ok=True)

        import requests  # not available in pyodide

        try:
            # download with tqdm adapted from:
            # https://github.com/shaypal5/tqdl/blob/189f7fd07f265d29af796bee28e0893e1396d237/tqdl/core.py
            # Streaming, so we can iterate over the response.
            r = requests.get(str(uri), stream=True)
            r.raise_for_status()
            # Total size in bytes.
            total_size = int(r.headers.get("content-length", 0))
            block_size = 1024  # 1 Kibibyte
            t = tqdm(total=total_size, unit="iB", unit_scale=True, desc=uri.path.split("/")[-1])
            tmp_path = local_path.with_suffix(f"{local_path.suffix}.part")
            with tmp_path.open("wb") as f:
                for data in r.iter_content(block_size):
                    t.update(len(data))
                    f.write(data)

            t.close()
            if total_size != 0 and hasattr(t, "n") and t.n != total_size:
                # todo: check more carefully and raise on real issue
                warnings.warn(f"Download ({t.n}) does not have expected size ({total_size}).")

            shutil.move(f.name, str(local_path))
        except Exception as e:
            raise RuntimeError(f"Failed to download {uri} ({e})")

    return local_path


T = typing.TypeVar("T")


def _resolve_json_from_url(
    url: str,
    expected_type: typing.Union[typing.Type[dict], typing.Type[T]] = dict,
    warning_msg: str = "Failed to fetch {url}: {error}",
    encoding: typing.Optional[str] = None,
) -> typing.Tuple[typing.Optional[T], typing.Optional[str]]:
    try:
        p = resolve_source(url)
        with p.open(encoding=encoding) as f:
            data = json.load(f)

        assert isinstance(data, expected_type)
    except Exception as e:
        data = None
        error: typing.Optional[str] = str(e)
        if warning_msg:
            warnings.warn(warning_msg.format(url=url, error=error))
    else:
        error = None

    return data, error


BIOIMAGEIO_SITE_CONFIG, BIOIMAGEIO_SITE_CONFIG_ERROR = _resolve_json_from_url(
    BIOIMAGEIO_SITE_CONFIG_URL, encoding="utf-8"
)
BIOIMAGEIO_COLLECTION, BIOIMAGEIO_COLLECTION_ERROR = _resolve_json_from_url(BIOIMAGEIO_COLLECTION_URL, encoding="utf-8")
if BIOIMAGEIO_COLLECTION is None:
    BIOIMAGEIO_COLLECTION_ENTRIES: typing.Optional[typing.Dict[str, typing.Tuple[str, str]]] = None
else:
    BIOIMAGEIO_COLLECTION_ENTRIES = {
        cr["id"]: (cr["type"], cr["rdf_source"])
        for cr in BIOIMAGEIO_COLLECTION.get("collection", [])
        if "id" in cr and "rdf_source" in cr and "type" in cr
    }
    # add resource versions explicitly
    BIOIMAGEIO_COLLECTION_ENTRIES.update(
        {
            f"{cr['id']}/{cv}": (
                cr["type"],
                cr["rdf_source"].replace(
                    f"/{cr['versions'][0]}", f"/{cv}"
                ),  # todo: improve this replace-version-monkeypatch
            )
            for cr in BIOIMAGEIO_COLLECTION.get("collection", [])
            for cv in cr.get("versions", [])
            if "id" in cr and "rdf_source" in cr and "type" in cr
        }
    )



import json
import sys
from argparse import ArgumentParser
from pathlib import Path
from urllib.request import urlretrieve

from bioimageio.spec.model import raw_nodes, schema
from bioimageio.spec.shared import get_args
from bioimageio.spec.shared.utils import resolve_bioimageio_descrcription

MANIFEST_URL = "https://raw.githubusercontent.com/bioimage-io/bioimage-io-models/gh-pages/manifest.bioimage.io.json"
WEIGHTS_FORMATS_OVERVIEW_PATH = (
    Path(__file__).parent / "../dist" / "weight_formats_spec.json"
)  # todo: weight_formats -> weights_formats

# defaults for transition period
consumer_defaults = {
    "ilastik": ["pytorch_state_dict", "onnx", "torchscript"],
    "zero": ["keras_hdf5"],
    "deepimagej": ["torchscript", "tensorflow_saved_model_bundle"],
    "imjoy": ["onnx"],
}


def parse_args():
    p = ArgumentParser(description=("script that generates weights formats overview"))
    p.add_argument("command", choices=["check", "generate"])

    args = p.parse_args()
    return args


def main(args):
    local = Path(urlretrieve(MANIFEST_URL)[0])
    with local.open() as f:
        collection = json.load(f)

    collection = collection["collections"]

    consumers = {c["id"]: c for c in collection}
    for consumer in consumers.values():
        if consumer["id"] in consumer_defaults:
            consumer["config"] = consumer.get("config", {})
            consumer["config"]["supported_weight_formats"] = consumer_defaults[consumer["id"]]

    weights_format_ids = get_args(raw_nodes.WeightsFormat)
    weights_format_class_names = [wf.title().replace("_", "") + "WeightsEntry" for wf in weights_format_ids]

    weights_formats = {
        wf: {
            "name": getattr(raw_nodes, wfcn).weights_format_name,
            "description": resolve_bioimageio_descrcription(getattr(schema, wfcn).bioimageio_description),
            "consumers": [
                cname
                for cname, c in consumers.items()
                if wf in c.get("supported_weights_formats", consumer_defaults.get(cname, []))
            ],
        }
        for wf, wfcn in zip(weights_format_ids, weights_format_class_names)
    }

    overview = {"consumers": consumers, "weight_formats": weights_formats}  # todo: weight_formats -> weights_formats

    if args.command == "generate":
        with WEIGHTS_FORMATS_OVERVIEW_PATH.open("w") as f:
            json.dump(overview, f, indent=4, sort_keys=True)
    elif args.command == "check":
        with WEIGHTS_FORMATS_OVERVIEW_PATH.open() as f:
            found = json.load(f)

        if found != overview:
            return 1
    else:
        raise NotImplementedError(args.command)


if __name__ == "__main__":
    args = parse_args()
    sys.exit(main(args))



import typing
from copy import deepcopy

import numpy
from marshmallow import RAISE, ValidationError, missing, pre_load, validates, validates_schema

from bioimageio.spec.dataset.v0_2.schema import Dataset as _Dataset
from bioimageio.spec.model.v0_3.schema import (
    KerasHdf5WeightsEntry as KerasHdf5WeightsEntry03,
    OnnxWeightsEntry as OnnxWeightsEntry03,
    Postprocessing as Postprocessing03,
    Preprocessing as Preprocessing03,
    TensorflowJsWeightsEntry as TensorflowJsWeightsEntry03,
    TensorflowSavedModelBundleWeightsEntry as TensorflowSavedModelBundleWeightsEntry03,
    _WeightsEntryBase as _WeightsEntryBase03,
    _common_sha256_hint,
)
from bioimageio.spec.rdf import v0_2 as rdf
from bioimageio.spec.shared import LICENSES, field_validators, fields
from bioimageio.spec.shared.common import get_args, get_args_flat
from bioimageio.spec.shared.schema import ImplicitOutputShape, ParametrizedInputShape, SharedBioImageIOSchema
from . import raw_nodes


class _BioImageIOSchema(SharedBioImageIOSchema):
    raw_nodes = raw_nodes


class _TensorBase(_BioImageIOSchema):
    name = fields.String(
        required=True,
        validate=field_validators.Predicate("isidentifier"),
        bioimageio_description="Tensor name. No duplicates are allowed.",
    )
    description = fields.String()
    axes = fields.Axes(
        required=True,
        bioimageio_description="""Axes identifying characters from: bitczyx. Same length and order as the axes in `shape`.

| character | description |
| --- | --- |
|  b  |  batch (groups multiple samples) |
|  i  |  instance/index/element |
|  t  |  time |
|  c  |  channel |
|  z  |  spatial dimension z |
|  y  |  spatial dimension y |
|  x  |  spatial dimension x |""",
    )
    data_type = fields.String(
        required=True,
        bioimageio_description="The data type of this tensor. For inputs, only `float32` is allowed and the consumer "
        "software needs to ensure that the correct data type is passed here. For outputs can be any of `float32, "
        "float64, (u)int8, (u)int16, (u)int32, (u)int64`. The data flow in bioimage.io models is explained "
        "[in this diagram.](https://docs.google.com/drawings/d/1FTw8-Rn6a6nXdkZ_SkMumtcjvur9mtIhRqLwnKqZNHM/edit).",
    )
    data_range = fields.Tuple(
        (fields.Float(allow_nan=True), fields.Float(allow_nan=True)),
        bioimageio_description="Tuple `(minimum, maximum)` specifying the allowed range of the data in this tensor. "
        "If not specified, the full data range that can be expressed in `data_type` is allowed.",
    )
    shape: fields.Union

    processing_name: str

    @validates_schema
    def validate_processing_kwargs(self, data, **kwargs):
        axes = data.get("axes", [])
        processing_list = data.get(self.processing_name, [])
        for processing in processing_list:
            kwargs = processing.kwargs or {}
            kwarg_axes = kwargs.get("axes", "")
            if any(a not in axes for a in kwarg_axes):
                raise ValidationError("`kwargs.axes` needs to be subset of axes")


class Preprocessing(Preprocessing03):
    kwargs = fields.Kwargs(
        bioimageio_description=f"Key word arguments as described in [preprocessing spec]"
        f"(https://github.com/bioimage-io/spec-bioimage-io/blob/gh-pages/preprocessing_spec_"
        f"{'_'.join(get_args(raw_nodes.FormatVersion)[-1].split('.')[:2])}.md)."
    )


class Postprocessing(Postprocessing03):
    kwargs = fields.Kwargs(
        bioimageio_description=f"Key word arguments as described in [postprocessing spec]"
        f"(https://github.com/bioimage-io/spec-bioimage-io/blob/gh-pages/postprocessing_spec_"
        f"{'_'.join(get_args(raw_nodes.FormatVersion)[-1].split('.')[:2])}.md)."
    )


class InputTensor(_TensorBase):
    shape = fields.Union(
        [
            fields.ExplicitShape(
                bioimageio_description="Exact shape with same length as `axes`, e.g. `shape: [1, 512, 512, 1]`"
            ),
            fields.Nested(
                ParametrizedInputShape(),
                bioimageio_description="A sequence of valid shapes given by `shape = min + k * step for k in {0, 1, ...}`.",
            ),
        ],
        required=True,
        bioimageio_description="Specification of input tensor shape.",
    )
    preprocessing = fields.List(
        fields.Nested(Preprocessing()), bioimageio_description="Description of how this input should be preprocessed."
    )
    processing_name = "preprocessing"

    @validates_schema
    def zero_batch_step_and_one_batch_size(self, data, **kwargs):
        axes = data.get("axes")
        shape = data.get("shape")

        if axes is None or shape is None:
            raise ValidationError("Failed to validate batch_step=0 and batch_size=1 due to other validation errors")

        axes = data["axes"]
        shape = data["shape"]

        bidx = axes.find("b")
        if bidx == -1:
            return

        if isinstance(shape, raw_nodes.ParametrizedInputShape):
            step = shape.step
            shape = shape.min

        elif isinstance(shape, list):
            step = [0] * len(shape)
        else:
            raise ValidationError(f"Unknown shape type {type(shape)}")

        if step[bidx] != 0:
            raise ValidationError(
                "Input shape step has to be zero in the batch dimension (the batch dimension can always be "
                "increased, but `step` should specify how to increase the minimal shape to find the largest "
                "single batch shape)"
            )

        if shape[bidx] != 1:
            raise ValidationError("Input shape has to be 1 in the batch dimension b.")


class OutputTensor(_TensorBase):
    shape = fields.Union(
        [
            fields.ExplicitShape(),
            fields.Nested(
                ImplicitOutputShape(),
                bioimageio_description="In reference to the shape of an input tensor, the shape of the output "
                "tensor is `shape = shape(input_tensor) * scale + 2 * offset`.",
            ),
        ],
        required=True,
        bioimageio_description="Specification of output tensor shape.",
    )
    halo = fields.List(
        fields.Integer(),
        bioimageio_description="The halo to crop from the output tensor (for example to crop away boundary effects or "
        "for tiling). The halo should be cropped from both sides, i.e. `shape_after_crop = shape - 2 * halo`. The "
        "`halo` is not cropped by the bioimage.io model, but is left to be cropped by the consumer software. Use "
        "`shape:offset` if the model output itself is cropped and input and output shapes not fixed.",
    )
    postprocessing = fields.List(
        fields.Nested(Postprocessing()),
        bioimageio_description="Description of how this output should be postprocessed.",
    )
    processing_name = "postprocessing"

    @validates_schema
    def matching_halo_length(self, data, **kwargs):
        shape = data.get("shape")
        halo = data.get("halo")
        if halo is None:
            return
        elif isinstance(shape, list) or isinstance(shape, raw_nodes.ImplicitOutputShape):
            if shape is None or len(halo) != len(shape):
                raise ValidationError(f"halo {halo} has to have same length as shape {shape}!")
        else:
            raise NotImplementedError(type(shape))


class _WeightsEntryBase(_WeightsEntryBase03):
    raw_nodes = raw_nodes
    dependencies = fields.Dependencies(
        bioimageio_description="Dependency manager and dependency file, specified as `<dependency manager>:<relative "
        "path to file>`. For example: 'conda:./environment.yaml', 'maven:./pom.xml', or 'pip:./requirements.txt'. "
        "These dependencies are only used for the specified weight format."
    )


class KerasHdf5WeightsEntry(KerasHdf5WeightsEntry03, _WeightsEntryBase):
    pass


class OnnxWeightsEntry(OnnxWeightsEntry03, _WeightsEntryBase):
    pass


class PytorchStateDictWeightsEntry(_WeightsEntryBase):
    bioimageio_description = "PyTorch state dictionary weights format"
    weights_format = fields.String(validate=field_validators.Equal("pytorch_state_dict"), required=True, load_only=True)
    architecture = fields.ImportableSource(
        required=True,
        bioimageio_description="Source code of the model architecture that either points to a "
        "local implementation: `<relative path to file>:<identifier of implementation within the file>` or the "
        "implementation in an available dependency: `<root-dependency>.<sub-dependency>.<identifier>`.\nFor example: "
        "`my_function.py:MyImplementation` or `bioimageio.core.some_module.some_class_or_function`.",
    )
    architecture_sha256 = fields.String(
        bioimageio_maybe_required=True,
        validate=field_validators.Length(equal=64),
        bioimageio_description="This field is only required if the architecture points to a source file. "
        "SHA256 checksum of the model source code file."
        + _common_sha256_hint.replace("    ", "        "),  # sha256 hint with one more intend level
    )
    kwargs = fields.Kwargs(
        bioimageio_description="Keyword arguments for the implementation specified by `architecture`."
    )
    pytorch_version = fields.Version()

    @validates_schema
    def sha_for_source_code_file(self, data, **kwargs):
        arch = data.get("architecture")
        if isinstance(arch, raw_nodes.ImportableModule):
            return
        elif isinstance(arch, raw_nodes.ImportableSourceFile):
            sha = data.get("architecture_sha256")
            if sha is None:
                raise ValidationError(
                    "When specifying 'architecture' with a callable from a source file, "
                    "the corresponding 'architecture_sha256' field is required."
                )


class TensorflowJsWeightsEntry(TensorflowJsWeightsEntry03, _WeightsEntryBase):
    pass


class TensorflowSavedModelBundleWeightsEntry(TensorflowSavedModelBundleWeightsEntry03, _WeightsEntryBase):
    pass


class Dataset(_Dataset):
    short_bioimageio_description = "in-place definition of [dataset RDF](https://github.com/bioimage-io/spec-bioimage-io/blob/gh-pages/dataset_spec_0_2.md)"


class TorchscriptWeightsEntry(_WeightsEntryBase):
    raw_nodes = raw_nodes

    bioimageio_description = "Torchscript weights format"
    weights_format = fields.String(validate=field_validators.Equal("torchscript"), required=True, load_only=True)
    pytorch_version = fields.Version()


WeightsEntry = typing.Union[
    KerasHdf5WeightsEntry,
    OnnxWeightsEntry,
    PytorchStateDictWeightsEntry,
    TensorflowJsWeightsEntry,
    TensorflowSavedModelBundleWeightsEntry,
    TorchscriptWeightsEntry,
]


class RunMode(_BioImageIOSchema):
    name = fields.String(required=True, bioimageio_description="The name of the `run_mode`")
    kwargs = fields.Kwargs()

    @validates("name")
    def warn_on_unrecognized_run_mode(self, value: str):
        if isinstance(value, str):
            self.warn("name", f"Unrecognized run mode '{value}'")


class LinkedDataset(_BioImageIOSchema):
    id = fields.String(bioimageio_description="dataset id")


class ModelParent(_BioImageIOSchema):
    id = fields.BioImageIO_ID(resource_type="model")
    uri = fields.Union(
        [fields.URI(), fields.Path()], bioimageio_description="URL or local relative path of a model RDF"
    )
    sha256 = fields.SHA256(bioimageio_description="Hash of the parent model RDF. Note: the hash is not validated")

    @validates_schema()
    def id_xor_uri(self, data, **kwargs):
        if ("id" in data) == ("uri" in data):
            raise ValidationError("Either 'id' or 'uri' are required (not both).")


class Model(rdf.schema.RDF):
    raw_nodes = raw_nodes

    class Meta:
        unknown = RAISE
        exclude = ("source",)  # while RDF does have a source field, Model does not

    bioimageio_description = f"""# BioImage.IO Model Resource Description File Specification {get_args(raw_nodes.FormatVersion)[-1]}
This specification defines the fields used in a BioImage.IO-compliant resource description file (`RDF`) for describing AI models with pretrained weights.
These fields are typically stored in YAML files which we call Model Resource Description Files or `model RDF`.
The model RDFs can be downloaded or uploaded to the bioimage.io website, produced or consumed by BioImage.IO-compatible consumers(e.g. image analysis software or other website).

The model RDF YAML file contains mandatory and optional fields. In the following description, optional fields are indicated by _optional_.
_optional*_ with an asterisk indicates the field is optional depending on the value in another field.
"""
    # todo: sync authors with RDF
    authors = fields.List(
        fields.Nested(rdf.schema.Author()),
        validate=field_validators.Length(min=1),
        required=True,
        bioimageio_description=rdf.schema.RDF.authors_bioimageio_description,
    )

    badges = missing  # todo: allow badges for Model (RDF has it)
    cite = fields.List(
        fields.Nested(rdf.schema.CiteEntry()),
        required=True,  # todo: unify authors with RDF (optional or required?)
        validate=field_validators.Length(min=1),
        bioimageio_description=rdf.schema.RDF.cite_bioimageio_description,
    )

    config = fields.YamlDict(
        bioimageio_description=rdf.schema.RDF.config_bioimageio_description
        + """
For example:
```yaml
config:
  # custom config for DeepImageJ, see https://github.com/bioimage-io/configuration/issues/23
  deepimagej:
    model_keys:
      # In principle the tag "SERVING" is used in almost every tf model
      model_tag: tf.saved_model.tag_constants.SERVING
      # Signature definition to call the model. Again "SERVING" is the most general
      signature_definition: tf.saved_model.signature_constants.DEFAULT_SERVING_SIGNATURE_DEF_KEY
    test_information:
      input_size: [2048x2048] # Size of the input images
      output_size: [1264x1264 ]# Size of all the outputs
      device: cpu # Device used. In principle either cpu or GPU
      memory_peak: 257.7 Mb # Maximum memory consumed by the model in the device
      runtime: 78.8s # Time it took to run the model
      pixel_size: [9.658E-4µmx9.658E-4µm] # Size of the pixels of the input
```
"""
    )

    documentation = fields.Union(
        [
            fields.URL(),
            fields.Path(
                validate=field_validators.Attribute(
                    "suffix",
                    field_validators.Equal(
                        ".md", error="{!r} is invalid; expected markdown file with '.md' extension."
                    ),
                )
            ),
        ],
        required=True,
        bioimageio_description="Relative path or URL to file with additional documentation in markdown. "
        "The file must be in markdown format with `.md` file name extension"
        "It is recommended to use `README.md` as the documentation name.",
    )

    format_version = fields.String(
        validate=field_validators.OneOf(get_args_flat(raw_nodes.FormatVersion)),
        required=True,
        bioimageio_description_order=0,
        bioimageio_description=f"""Version of the BioImage.IO Model Resource Description File Specification used.
This is mandatory, and important for the consumer software to verify before parsing the fields.
The recommended behavior for the implementation is to keep backward compatibility and throw an error if the model yaml
is in an unsupported format version. The current format version described here is
{get_args(raw_nodes.FormatVersion)[-1]}""",
    )

    git_repo = fields.URL(
        bioimageio_description=rdf.schema.RDF.git_repo_bioimageio_description
        + "If the model is contained in a subfolder of a git repository, then a url to the exact folder"
        + "(which contains the configuration yaml file) should be used."
    )

    inputs = fields.List(
        fields.Nested(InputTensor()),
        validate=field_validators.Length(min=1),
        required=True,
        bioimageio_description="Describes the input tensors expected by this model.",
    )

    @validates("inputs")
    def no_duplicate_input_tensor_names(self, value: typing.List[raw_nodes.InputTensor]):
        if not isinstance(value, list) or not all(isinstance(v, raw_nodes.InputTensor) for v in value):
            raise ValidationError("Could not check for duplicate input tensor names due to another validation error.")

        names = [t.name for t in value]
        if len(names) > len(set(names)):
            raise ValidationError("Duplicate input tensor names are not allowed.")

    license = fields.String(
        validate=field_validators.OneOf(LICENSES),
        required=True,
        bioimageio_description=rdf.schema.RDF.license_bioimageio_description,
    )

    name = fields.String(
        required=True,
        bioimageio_description="Name of this model. It should be human-readable and only contain letters, numbers, "
        "underscore '_', minus '-' or spaces and not be longer than 64 characters.",
    )

    outputs = fields.List(
        fields.Nested(OutputTensor()),
        validate=field_validators.Length(min=1),
        bioimageio_description="Describes the output tensors from this model.",
    )

    @validates("outputs")
    def no_duplicate_output_tensor_names(self, value: typing.List[raw_nodes.OutputTensor]):
        if not isinstance(value, list) or not all(isinstance(v, raw_nodes.OutputTensor) for v in value):
            raise ValidationError("Could not check for duplicate output tensor names due to another validation error.")

        names = [t["name"] if isinstance(t, dict) else t.name for t in value]
        if len(names) > len(set(names)):
            raise ValidationError("Duplicate output tensor names are not allowed.")

    @validates_schema
    def inputs_and_outputs(self, data, **kwargs):
        ipts: typing.List[raw_nodes.InputTensor] = data.get("inputs")
        outs: typing.List[raw_nodes.OutputTensor] = data.get("outputs")
        if any(
            [
                not isinstance(ipts, list),
                not isinstance(outs, list),
                not all(isinstance(v, raw_nodes.InputTensor) for v in ipts),
                not all(isinstance(v, raw_nodes.OutputTensor) for v in outs),
            ]
        ):
            raise ValidationError("Could not check for duplicate tensor names due to another validation error.")

        # no duplicate tensor names
        names = [t.name for t in ipts + outs]  # type: ignore
        if len(names) > len(set(names)):
            raise ValidationError("Duplicate tensor names are not allowed.")

        tensors_by_name: typing.Dict[str, typing.Union[raw_nodes.InputTensor, raw_nodes.OutputTensor]] = {
            t.name: t for t in ipts + outs  # type: ignore
        }

        # minimum shape leads to valid output:
        # output with subtracted halo has to result in meaningful output even for the minimal input
        # see https://github.com/bioimage-io/spec-bioimage-io/issues/392
        def get_min_shape(t) -> numpy.ndarray:
            if isinstance(t.shape, raw_nodes.ParametrizedInputShape):
                shape = numpy.array(t.shape.min)
            elif isinstance(t.shape, raw_nodes.ImplicitOutputShape):
                shape = get_min_shape(tensors_by_name[t.shape.reference_tensor]) * t.shape.scale + 2 * numpy.array(
                    t.shape.offset
                )
            else:
                shape = numpy.array(t.shape)

            return shape

        for out in outs:
            if isinstance(out.shape, raw_nodes.ImplicitOutputShape) and len(out.shape) != len(
                tensors_by_name[out.shape.reference_tensor].shape
            ):
                raise ValidationError(
                    f"Referenced tensor {out.shape.reference_tensor} "
                    f"with {len(tensors_by_name[out.shape.reference_tensor].shape)} dimensions does not match "
                    f"output tensor {out.name} with {len(out.shape)} dimensions."
                )

            min_out_shape = get_min_shape(out)
            if out.halo:
                halo = out.halo
                halo_msg = f" for halo {out.halo}"
            else:
                halo = [0] * len(min_out_shape)
                halo_msg = ""

            if any([s - 2 * h < 1 for s, h in zip(min_out_shape, halo)]):
                raise ValidationError(f"Minimal shape {min_out_shape} of output {out.name} is too small{halo_msg}.")

    packaged_by = fields.List(
        fields.Nested(rdf.schema.Author()),
        bioimageio_description=f"The persons that have packaged and uploaded this model. Only needs to be specified if "
        f"different from `authors` in root or any entry in `weights`.",
    )

    parent = fields.Nested(
        ModelParent(),
        bioimageio_description="The model from which this model is derived, e.g. by fine-tuning the weights.",
    )

    run_mode = fields.Nested(
        RunMode(),
        bioimageio_description="Custom run mode for this model: for more complex prediction procedures like test time "
        "data augmentation that currently cannot be expressed in the specification. "
        "No standard run modes are defined yet.",
    )

    sample_inputs = fields.List(
        fields.Union([fields.URI(), fields.Path()]),
        validate=field_validators.Length(min=1),
        bioimageio_description="List of URIs/local relative paths to sample inputs to illustrate possible inputs for "
        "the model, for example stored as png or tif images. "
        "The model is not tested with these sample files that serve to inform a human user about an example use case.",
    )
    sample_outputs = fields.List(
        fields.Union([fields.URI(), fields.Path()]),
        validate=field_validators.Length(min=1),
        bioimageio_description="List of URIs/local relative paths to sample outputs corresponding to the "
        "`sample_inputs`.",
    )

    test_inputs = fields.List(
        fields.Union([fields.URI(), fields.Path()]),
        validate=field_validators.Length(min=1),
        required=True,
        bioimageio_description="List of URIs or local relative paths to test inputs as described in inputs for "
        "**a single test case**. "
        "This means if your model has more than one input, you should provide one URI for each input."
        "Each test input should be a file with a ndarray in "
        "[numpy.lib file format](https://numpy.org/doc/stable/reference/generated/numpy.lib.format.html#module-numpy.lib.format)."
        "The extension must be '.npy'.",
    )
    test_outputs = fields.List(
        fields.Union([fields.URI(), fields.Path()]),
        validate=field_validators.Length(min=1),
        required=True,
        bioimageio_description="Analog to test_inputs.",
    )

    timestamp = fields.DateTime(
        required=True,
        bioimageio_description="Timestamp of the initial creation of this model in [ISO 8601]"
        "(#https://en.wikipedia.org/wiki/ISO_8601) format.",
    )

    training_data = fields.Union([fields.Nested(Dataset()), fields.Nested(LinkedDataset())])

    weights = fields.Dict(
        fields.String(
            validate=field_validators.OneOf(get_args(raw_nodes.WeightsFormat)),
            required=True,
            bioimageio_description="Format of this set of weights. "
            f"One of: {', '.join(get_args(raw_nodes.WeightsFormat))}",
        ),
        fields.Union(
            [fields.Nested(we()) for we in get_args(WeightsEntry)],
            short_bioimageio_description=(
                "The weights for this model. Weights can be given for different formats, but should "
                "otherwise be equivalent. "
                "See [weight_formats_spec_0_4.md]"
                "(https://github.com/bioimage-io/spec-bioimage-io/blob/gh-pages/weight_formats_spec_0_4.md) "
                "for the required and optional fields per weight format. "
                "The available weight formats determine which consumers can use this model."
            ),
        ),
        required=True,
    )

    @pre_load
    def add_weights_format_key_to_weights_entry_value(self, data: dict, many=False, partial=False, **kwargs):
        data = deepcopy(data)  # Schema.validate() calls pre_load methods, thus we should not modify the input data
        if many or partial:
            raise NotImplementedError

        for weights_format, weights_entry in data.get("weights", {}).items():
            if "weights_format" in weights_entry:
                raise ValidationError(f"Got unexpected key 'weights_format' in weights entry {weights_format}")

            weights_entry["weights_format"] = weights_format

        return data

    @validates_schema
    def validate_reference_tensor_names(self, data, **kwargs):
        def get_tnames(tname: str):
            return [t.get("name") if isinstance(t, dict) else t.name for t in data.get(tname, [])]

        valid_input_tensor_references = get_tnames("inputs")
        ins = data.get("inputs", [])
        outs = data.get("outputs", [])
        if not isinstance(ins, list) or not isinstance(outs, list):
            raise ValidationError(
                f"Failed to validate reference tensor names due to other validation errors in inputs/outputs."
            )

        for t in outs:
            if not isinstance(t, raw_nodes.OutputTensor):
                raise ValidationError("Failed to validate reference tensor names due to validation errors in outputs")

            if t.postprocessing is missing:
                continue

            for postpr in t.postprocessing:
                if postpr.kwargs is missing:
                    continue

                ref_tensor = postpr.kwargs.get("reference_tensor", missing)
                if ref_tensor is not missing and ref_tensor not in valid_input_tensor_references:
                    raise ValidationError(f"{ref_tensor} not found in inputs")

        for t in ins:
            if not isinstance(t, raw_nodes.InputTensor):
                raise ValidationError("Failed to validate reference tensor names due to validation errors in inputs")

            if t.preprocessing is missing:
                continue

            for prep in t.preprocessing:
                if prep.kwargs is missing:
                    continue

                ref_tensor = prep.kwargs.get("reference_tensor", missing)
                if ref_tensor is not missing and ref_tensor not in valid_input_tensor_references:
                    raise ValidationError(f"{ref_tensor} not found in inputs")

                if ref_tensor == t.name:
                    raise ValidationError(f"invalid self reference for preprocessing of tensor {t.name}")

    @validates_schema
    def weights_entries_match_weights_formats(self, data, **kwargs):
        weights: typing.Dict[str, WeightsEntry] = data.get("weights", {})
        for weights_format, weights_entry in weights.items():
            if not isinstance(weights_entry, get_args(raw_nodes.WeightsEntry)):
                raise ValidationError("Cannot validate keys in weights field due to other validation errors.")

            if weights_format in ["pytorch_state_dict", "torchscript"]:
                if weights_format == "pytorch_state_dict":
                    assert isinstance(weights_entry, raw_nodes.PytorchStateDictWeightsEntry)
                elif weights_format == "torchscript":
                    assert isinstance(weights_entry, raw_nodes.TorchscriptWeightsEntry)
                else:
                    raise NotImplementedError

                if weights_entry.dependencies is missing and weights_entry.pytorch_version is missing:
                    self.warn(f"weights:{weights_format}", "missing 'pytorch_version'")

            if weights_format in ["keras_hdf5", "tensorflow_js", "tensorflow_saved_model_bundle"]:
                if weights_format == "keras_hdf5":
                    assert isinstance(weights_entry, raw_nodes.KerasHdf5WeightsEntry)
                elif weights_format == "tensorflow_js":
                    assert isinstance(weights_entry, raw_nodes.TensorflowJsWeightsEntry)
                elif weights_format == "tensorflow_saved_model_bundle":
                    assert isinstance(weights_entry, raw_nodes.TensorflowSavedModelBundleWeightsEntry)
                else:
                    raise NotImplementedError

                if weights_entry.dependencies is missing and weights_entry.tensorflow_version is missing:
                    self.warn(f"weights:{weights_format}", "missing 'tensorflow_version'")

            if weights_format == "onnx":
                assert isinstance(weights_entry, raw_nodes.OnnxWeightsEntry)
                if weights_entry.dependencies is missing and weights_entry.opset_version is missing:
                    self.warn(f"weights:{weights_format}", "missing 'opset_version'")



import ast
import sys
import warnings
from typing import Callable, Union
from urllib.parse import urlparse

try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal  # type: ignore


def snake_case_to_camel_case(string: str) -> str:
    return "".join([s.title() for s in string.split("_")])


def get_ref_url(type_: Literal["class", "function"], name: str, github_file_url: str) -> str:
    """get github url with line range fragment to reference implementation from non-raw github file url

    example:
    >>> get_ref_url(
        "class",
        "Binarize",
        "https://github.com/bioimage-io/core-bioimage-io-python/blob/main/bioimageio/core/prediction_pipeline/_processing.py"
    )
    https://github.com/bioimage-io/core-bioimage-io-python/blob/main/bioimageio/core/prediction_pipeline/_processing.py#L107-L112
    """
    # hotfix to handle requests not available in pyodide, see
    # https://github.com/bioimage-io/bioimage.io/issues/216#issuecomment-1012422194
    try:
        import requests  # not available in pyodide
    except Exception:
        warnings.warn(f"Could not reslove {github_file_url} because requests library is not available.")
        return "URL NOT RESOLVED"

    assert not urlparse(github_file_url).fragment, "unexpected url fragment"
    look_for = {"class": ast.ClassDef, "function": ast.FunctionDef}[type_]
    raw_github_file_url = github_file_url.replace("github.com", "raw.githubusercontent.com").replace("/blob/", "/")
    try:
        code = requests.get(raw_github_file_url).text
    except requests.RequestException as e:
        warnings.warn(f"Could not resolve {github_file_url} due to {e}. Please check your internet connection.")
        return "URL NOT RESOLVED"
    tree = ast.parse(code)

    for d in tree.body:
        if isinstance(d, look_for):
            assert hasattr(d, "name")
            if d.name == name:  # type: ignore
                assert hasattr(d, "decorator_list")
                start = d.decorator_list[0].lineno if d.decorator_list else d.lineno  # type: ignore
                if sys.version_info >= (3, 8):
                    stop = d.end_lineno
                else:
                    stop = d.lineno + 1
                break
    else:
        raise ValueError(f"{type_} {name} not found in {github_file_url}")

    return f"{github_file_url}#L{start}-L{stop}"


def resolve_bioimageio_descrcription(descr: Union[Callable[[], str], str]):
    if callable(descr):
        return descr()
    else:
        return descr



from marshmallow import EXCLUDE, ValidationError, validates, validates_schema

from bioimageio.spec.shared import (
    BIOIMAGEIO_SITE_CONFIG,
    BIOIMAGEIO_SITE_CONFIG_ERROR,
    LICENSES,
    field_validators,
    fields,
)
from bioimageio.spec.shared.common import get_args, get_patched_format_version
from bioimageio.spec.shared.schema import SharedBioImageIOSchema, WithUnknown
from bioimageio.spec.shared.utils import is_valid_orcid_id
from . import raw_nodes
from .raw_nodes import FormatVersion


class _BioImageIOSchema(SharedBioImageIOSchema):
    raw_nodes = raw_nodes


class Attachments(_BioImageIOSchema, WithUnknown):
    files = fields.List(
        fields.Union([fields.URI(), fields.Path()]),
        bioimageio_description="File attachments; included when packaging the resource.",
    )


class _Person(_BioImageIOSchema):
    name = fields.String(bioimageio_description="Full name.")
    affiliation = fields.String(bioimageio_description="Affiliation.")
    email = fields.Email()
    github_user = fields.String(bioimageio_description="GitHub user name.")  # todo: add validation?
    orcid = fields.String(
        validate=[
            field_validators.Length(19),
            lambda oid: all(oid[idx] == "-" for idx in [4, 9, 14]),
            lambda oid: is_valid_orcid_id(oid.replace("-", "")),
        ],
        bioimageio_description="[orcid](https://support.orcid.org/hc/en-us/sections/360001495313-What-is-ORCID) id "
        "in hyphenated groups of 4 digits, e.g. '0000-0001-2345-6789' (and [valid]("
        "https://support.orcid.org/hc/en-us/articles/360006897674-Structure-of-the-ORCID-Identifier"
        ") as per ISO 7064 11,2.)",
    )


class Author(_Person):
    name = fields.String(require=True, bioimageio_description="Full name.")


class Maintainer(_Person):
    github_user = fields.String(require=True, bioimageio_description="GitHub user name.")


class Badge(_BioImageIOSchema):
    bioimageio_description = "Custom badge."
    label = fields.String(required=True, bioimageio_description="e.g. 'Open in Colab'")
    icon = fields.String(bioimageio_description="e.g. 'https://colab.research.google.com/assets/colab-badge.svg'")
    url = fields.Union(
        [fields.URL(), fields.Path()],
        bioimageio_description="e.g. 'https://colab.research.google.com/github/HenriquesLab/ZeroCostDL4Mic/blob/master/Colab_notebooks/U-net_2D_ZeroCostDL4Mic.ipynb'",
    )


class CiteEntry(_BioImageIOSchema):
    text = fields.String(required=True, bioimageio_description="free text description")
    doi = fields.DOI(
        bioimageio_maybe_required=True, bioimageio_description="digital object identifier, see https://www.doi.org/"
    )
    url = fields.String(bioimageio_maybe_required=True)  # todo: change to fields.URL

    @validates_schema
    def doi_or_url(self, data, **kwargs):
        if data.get("doi") is None and data.get("url") is None:
            raise ValidationError("doi or url needs to be specified in a citation")


class RDF(_BioImageIOSchema):
    class Meta:
        unknown = EXCLUDE

    bioimageio_description = f"""# BioImage.IO Resource Description File Specification {get_args(FormatVersion)[-1]}
This specification defines the fields used in a general BioImage.IO-compliant resource description file (`RDF`).
An RDF is stored as a YAML file and describes resources such as models, datasets, applications and notebooks. 
Note that models are described with an extended Model RDF specification.

The RDF contains mandatory and optional fields. In the following description, optional fields are indicated by 
_optional_. _optional*_ with an asterisk indicates the field is optional depending on the value in another field.
If no specialized RDF exists for the specified type (like model RDF for type='model') additional fields may be 
specified.
"""

    attachments = fields.Nested(Attachments(), bioimageio_description="Additional unknown keys are allowed.")

    authors_bioimageio_description = (
        "A list of authors. The authors are the creators of the specifications and the primary points of contact."
    )
    authors = fields.List(fields.Nested(Author()), bioimageio_description=authors_bioimageio_description)

    badges = fields.List(fields.Nested(Badge()), bioimageio_description="a list of badges")

    cite_bioimageio_description = """A list of citation entries.
Each entry contains a mandatory `text` field and either one or both of `doi` and `url`.
E.g. the citation for the model architecture and/or the training data used."""
    cite = fields.List(fields.Nested(CiteEntry()), bioimageio_description=cite_bioimageio_description)

    config_bioimageio_description = (
        "A custom configuration field that can contain any keys not present in the RDF spec. "
        "This means you should not store, for example, github repo URL in `config` since we already have the "
        "`git_repo` key defined in the spec.\n"
        "Keys in `config` may be very specific to a tool or consumer software. To avoid conflicted definitions, "
        "it is recommended to wrap configuration into a sub-field named with the specific domain or tool name, "
        """for example:

```yaml
   config:
      bioimage_io:  # here is the domain name
        my_custom_key: 3837283
        another_key:
           nested: value
      imagej:
        macro_dir: /path/to/macro/file
```
"""
        "If possible, please use [`snake_case`](https://en.wikipedia.org/wiki/Snake_case) for keys in `config`."
    )
    config = fields.YamlDict(bioimageio_descriptio=config_bioimageio_description)

    covers = fields.List(
        fields.Union([fields.URL(), fields.Path()]),
        bioimageio_description="A list of cover images provided by either a relative path to the model folder, or a "
        "hyperlink starting with 'http[s]'. Please use an image smaller than 500KB and an aspect ratio width to height "
        "of 2:1. The supported image formats are: 'jpg', 'png', 'gif'.",  # todo: field_validators image format
    )

    description = fields.String(required=True, bioimageio_description="A string containing a brief description.")

    documentation = fields.Union(
        [
            fields.URL(),
            fields.Path(
                validate=field_validators.Attribute(
                    "suffix",
                    field_validators.Equal(
                        ".md", error="{!r} is invalid; expected markdown file with '.md' extension."
                    ),
                )
            ),
        ],
        bioimageio_description="URL or relative path to markdown file with additional documentation. "
        "For markdown files the recommended documentation file name is `README.md`.",
    )

    download_url = fields.URL(bioimageio_description="optional url to download the resource from")

    format_version = fields.String(
        required=True,
        bioimageio_description_order=0,
        bioimageio_description=(
            "Version of the BioImage.IO Resource Description File Specification used."
            f"The current general format version described here is {get_args(FormatVersion)[-1]}. "
            "Note: The general RDF format is not to be confused with specialized RDF format like the Model RDF format."
        ),
    )

    @validates_schema
    def format_version_matches_type(self, data, **kwargs):
        format_version = data.get("format_version")
        type_ = data.get("type")
        try:
            patched_format_version = get_patched_format_version(type_, format_version)
            if format_version.split(".") > patched_format_version.split("."):
                raise ValueError(
                    f"Unknown format_version {format_version} (latest patch: {patched_format_version}; latest format version: )"
                )
        except Exception as e:
            raise ValidationError(f"Invalid format_version {format_version} for RDF type {type_}. (error: {e})")

    git_repo_bioimageio_description = "A url to the git repository, e.g. to Github or Gitlab."
    git_repo = fields.URL(bioimageio_description=git_repo_bioimageio_description)

    icon = fields.String(
        bioimageio_description="an icon for the resource"
    )  # todo: limit length? validate=field_validators.Length(max=1)

    id = fields.String(bioimageio_description="Unique id within a collection of resources.")
    license_bioimageio_description = (
        "A [SPDX license identifier](https://spdx.org/licenses/)(e.g. `CC-BY-4.0`, `MIT`, "
        "`BSD-2-Clause`). We don't support custom license beyond the SPDX license list, if you need that please send "
        "an Github issue to discuss your intentions with the community."
    )
    license = fields.String(  # todo: make mandatory?
        # validate=field_validators.OneOf(LICENSES),  # enforce license id
        bioimageio_description=license_bioimageio_description
    )

    @validates("license")
    def warn_about_deprecated_spdx_license(self, value: str):
        license_info = LICENSES.get(value)
        if license_info is None:
            self.warn("license", f"{value} is not a recognized SPDX license identifier. See https://spdx.org/licenses/")
        else:
            if license_info.get("isDeprecatedLicenseId", False):
                self.warn("license", f"{value} ({license_info['name']}) is deprecated.")

            if not license_info.get("isFsfLibre", False):
                self.warn("license", f"{value} ({license_info['name']}) is not FSF Free/libre.")

    links = fields.List(fields.String(), bioimageio_description="links to other bioimage.io resources")

    maintainers = fields.List(fields.Nested(Maintainer()), bioimageio_description="Maintainers of this resource.")

    name = fields.String(required=True, bioimageio_description="name of the resource, a human-friendly name")

    @validates
    def warn_about_long_name(self, value: str):
        if isinstance(value, str):
            if len(value) > 64:
                self.warn(
                    "name", f"Length of name ({len(value)}) exceeds the recommended maximum length of 64 characters."
                )
        else:
            self.warn("name", f"Could not check length of name {value}.")

    rdf_source = fields.Union(
        [fields.URL(), fields.DOI()], bioimageio_description="url or doi to the source of the resource definition"
    )
    source = fields.Union(
        [fields.URI(), fields.Path()],
        bioimageio_description="url or local relative path to the source of the resource",
    )

    tags = fields.List(fields.String(), bioimageio_description="A list of tags.")

    @validates("tags")
    def warn_about_tag_categories(self, value):
        if BIOIMAGEIO_SITE_CONFIG is None:
            error = BIOIMAGEIO_SITE_CONFIG_ERROR
        else:
            missing_categories = []
            try:
                categories = {
                    c["type"]: c.get("tag_categories", {}) for c in BIOIMAGEIO_SITE_CONFIG["resource_categories"]
                }.get(self.__class__.__name__.lower(), {})
                for cat, entries in categories.items():
                    if not any(e in value for e in entries):
                        missing_categories.append({cat: entries})
            except Exception as e:
                error = str(e)
            else:
                error = None
                if missing_categories:
                    self.warn("tags", f"Missing tags for categories: {missing_categories}")

        if error is not None:
            self.warn("tags", f"could not check tag categories ({error})")

    type = fields.String(required=True)

    # todo: restrict valid RDF types?
    @validates("type")
    def validate_type(self, value):
        schema_type = self.__class__.__name__.lower()
        if value != schema_type:
            self.warn("type", f"Unrecognized type '{value}'. Validating as {schema_type}.")

    version = fields.Version(
        bioimageio_description="The version number of the model. The version number format must be a string in "
        "`MAJOR.MINOR.PATCH` format following the guidelines in Semantic Versioning 2.0.0 (see https://semver.org/), "
        "e.g. the initial version number should be `0.1.0`."
    )



import copy
from typing import Any, Dict


def maybe_convert(data: Dict[str, Any]) -> Dict[str, Any]:
    data = copy.deepcopy(data)
    if data.get("format_version") in ("0.2.0", "0.2.1"):
        # move all type groups to the 'collection' field
        if "collection" not in data:
            data["collection"] = []

        for group in ["application", "model", "dataset", "notebook"]:
            if group in data:
                data["collection"] += data[group]
                data["collection"][-1]["type"] = group

        config = data.get("config")
        if config and isinstance(config, dict):
            id_ = config.pop("id", data.get("id"))
            if id_ is not None:
                data["id"] = id_

        data["format_version"] = "0.2.2"

    return data



import json
from pathlib import Path
from setuptools import find_namespace_packages, setup

# Get the long description from the README file
ROOT_DIR = Path(__file__).parent.resolve()
long_description = (ROOT_DIR / "README.md").read_text(encoding="utf-8")
VERSION_FILE = ROOT_DIR / "bioimageio" / "spec" / "VERSION"
VERSION = json.loads(VERSION_FILE.read_text(encoding="utf-8"))["version"]


setup(
    name="bioimageio.spec",
    version=VERSION,
    description="Parser and validator library for BioImage.IO specifications",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/bioimage-io/spec-bioimage-io",
    author="Bioimage Team",
    classifiers=[  # Optional
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
    ],
    packages=find_namespace_packages(exclude=["tests"]),  # Required
    install_requires=[
        "marshmallow-jsonschema",
        "marshmallow-union",
        "marshmallow>=3.6.0,<4.0",
        "numpy",
        "packaging>=17.0",
        "requests;platform_system!='Emscripten'",
        "ruamel.yaml;platform_system!='Emscripten'",
        "tqdm;platform_system!='Emscripten'",
        "typer",
        "typing-extensions",
    ],
    entry_points={"console_scripts": ["bioimageio = bioimageio.spec.__main__:app"]},
    extras_require={"test": ["pytest", "tox", "mypy"], "dev": ["pre-commit"]},
    scripts=[
        "scripts/generate_json_specs.py",
        "scripts/generate_processing_docs.py",
        "scripts/generate_rdf_docs.py",
        "scripts/generate_weight_formats_docs.py",
    ],
    include_package_data=True,
    project_urls={  # Optional
        "Bug Reports": "https://github.com/bioimage-io/spec-bioimage-io/issues",
        "Source": "https://github.com/bioimage-io/spec-bioimage-io",
    },
)



import copy
from typing import Any, Dict

from marshmallow import missing


def convert_model_from_v0_3_to_0_4_0(data: Dict[str, Any]) -> Dict[str, Any]:
    from bioimageio.spec.model import v0_3

    data = copy.deepcopy(data)

    data = v0_3.converters.maybe_convert(data)
    v0_3.schema.Model().validate(data)

    data.pop("language", None)
    data.pop("framework", None)

    architecture = data.pop("source", missing)
    architecture_sha256 = data.pop("sha256", missing)
    kwargs = data.pop("kwargs", missing)
    pytorch_state_dict_weights_entry = data.get("weights", {}).get("pytorch_state_dict")
    if pytorch_state_dict_weights_entry is not None:
        if architecture is not missing:
            pytorch_state_dict_weights_entry["architecture"] = architecture

        if architecture_sha256 is not missing:
            pytorch_state_dict_weights_entry["architecture_sha256"] = architecture_sha256

        if kwargs is not missing:
            pytorch_state_dict_weights_entry["kwargs"] = kwargs

    torchscript_weights_entry = data.get("weights", {}).pop("pytorch_script", None)
    if torchscript_weights_entry is not None:
        data["weights"]["torchscript"] = torchscript_weights_entry

    data["format_version"] = "0.4.0"

    return data


def convert_model_from_v0_4_0_to_0_4_1(data: Dict[str, Any]) -> Dict[str, Any]:
    data = dict(data)

    # move dependencies from root to pytorch_state_dict weights entry
    deps = data.pop("dependencies", None)
    weights = data.get("weights", {})
    if deps and weights and isinstance(weights, dict):
        entry = weights.get("pytorch_state_dict")
        if entry and isinstance(entry, dict):
            entry["dependencies"] = deps

    data["format_version"] = "0.4.1"
    return data


def convert_model_from_v0_4_4_to_0_4_5(data: Dict[str, Any]) -> Dict[str, Any]:
    data = dict(data)

    parent = data.pop("parent", None)
    if parent and "uri" in parent:
        data["parent"] = parent["uri"]

    data["format_version"] = "0.4.5"
    return data


def maybe_convert(data: Dict[str, Any]) -> Dict[str, Any]:
    """auto converts model 'data' to newest format"""
    major, minor, patch = map(int, data.get("format_version", "0.3.0").split("."))
    if major == 0 and minor < 4:
        data = convert_model_from_v0_3_to_0_4_0(data)

    if data["format_version"] == "0.4.0":
        data = convert_model_from_v0_4_0_to_0_4_1(data)

    if data["format_version"] in ("0.4.1", "0.4.2", "0.4.3"):
        data["format_version"] = "0.4.4"

    if data["format_version"] == "0.4.4":
        data = convert_model_from_v0_4_4_to_0_4_5(data)

    # remove 'future' from config if no other than the used future entries exist
    config = data.get("config", {})
    if config.get("future") == {}:
        del config["future"]

    # remove 'config' if now empty
    if data.get("config") == {}:
        del data["config"]

    return data



from __future__ import annotations

import json
from pathlib import Path

from ._resolve_source import (
    BIOIMAGEIO_COLLECTION,
    BIOIMAGEIO_COLLECTION_ERROR,
    BIOIMAGEIO_SITE_CONFIG,
    BIOIMAGEIO_SITE_CONFIG_ERROR,
    RDF_NAMES,
    _resolve_json_from_url,
    resolve_local_source,
    resolve_rdf_source,
    resolve_rdf_source_and_type,
    resolve_source,
    source_available,
)
from ._update_nested import update_nested
from .common import get_args, yaml  # noqa

_license_file = Path(__file__).parent.parent / "static" / "licenses.json"
_license_data = json.loads(_license_file.read_text(encoding="utf-8"))

LICENSES = {x["licenseId"]: x for x in _license_data["licenses"]}
LICENSE_DATA_VERSION = _license_data["licenseListVersion"]



import copy
from typing import Any, Dict


def maybe_convert(data: Dict[str, Any]) -> Dict[str, Any]:
    data = copy.deepcopy(data)

    # we unofficially accept strings as author entries...
    authors = data.get("authors")
    if isinstance(authors, list):
        data["authors"] = [{"name": a} if isinstance(a, str) else a for a in authors]

    if data.get("format_version") in ("0.2.0", "0.2.1"):
        data["format_version"] = "0.2.2"

    return data



import os
import pathlib
import tempfile
import warnings
from typing import Any, Generic, Iterable, List, Optional, Sequence

try:
    from typing import Literal, get_args, get_origin, Protocol
except ImportError:
    from typing_extensions import Literal, get_args, get_origin, Protocol  # type: ignore


try:
    from ruamel.yaml import YAML  # not available in pyodide
except ImportError:
    yaml: Optional["MyYAML"] = None
else:

    class MyYAML(YAML):
        """add convenient improvements over YAML
        improve dump:
            - make sure to dump with utf-8 encoding. on windows encoding 'windows-1252' may otherwise be used
            - expose indentation kwargs for dump
        """

        def dump(self, data, stream=None, *, transform=None):
            if isinstance(stream, pathlib.Path):
                with stream.open("wt", encoding="utf-8") as f:
                    return super().dump(data, f, transform=transform)
            else:
                return super().dump(data, stream, transform=transform)

    yaml = MyYAML(typ="safe")


try:
    from tqdm import tqdm  # not available in pyodide
except ImportError:

    class tqdm:  # type: ignore
        """no-op tqdm"""

        def __init__(self, iterable: Optional[Iterable] = None, *args, **kwargs):
            self.iterable = iterable

        def __iter__(self):
            if self.iterable is not None:
                yield from self.iterable

        def update(self, *args, **kwargs):
            pass


BIOIMAGEIO_CACHE_PATH = pathlib.Path(
    os.getenv("BIOIMAGEIO_CACHE_PATH", pathlib.Path(tempfile.gettempdir()) / "bioimageio_cache")
)
BIOIMAGEIO_USE_CACHE = os.getenv("BIOIMAGEIO_USE_CACHE", "true").lower() in ("true", "yes", "1")

# keep a reference to temporary directories and files.
# These temporary locations are used instead of paths in BIOIMAGEIO_CACHE_PATH if BIOIMAGEIO_USE_CACHE is true,
# so that we guarantee that these temporary directories/file objects are not garbage collected too early
# and thus their content removed from disk, while we still have a pathlib.Path pointing there
no_cache_tmp_list: List[Any] = []

BIOIMAGEIO_SITE_CONFIG_URL = "https://raw.githubusercontent.com/bioimage-io/bioimage.io/main/site.config.json"
BIOIMAGEIO_COLLECTION_URL = "https://bioimage-io.github.io/collection-bioimage-io/collection.json"


DOI_REGEX = r"^10[.][0-9]{4,9}\/[-._;()\/:A-Za-z0-9]+$"
RDF_NAMES = ("rdf.yaml", "model.yaml")


class ValidationWarning(UserWarning):
    """a warning category to warn with during RDF validation"""

    @staticmethod
    def get_warning_summary(val_warns: Sequence[warnings.WarningMessage]) -> dict:
        """Summarize warning messages of the ValidationWarning category"""

        def add_to_summary(s, keys, msg):
            key = keys.pop(0)
            if "[" in key:
                key, rest = key.split("[")
                assert rest[-1] == "]"
                idx = int(rest[:-1])
            else:
                idx = None

            if key not in s:
                s[key] = {} if keys or idx is not None else msg

            s = s[key]

            if idx is not None:
                if idx not in s:
                    s[idx] = {} if keys else {"warning": msg}

                s = s[idx]

            if keys:
                assert isinstance(s, dict), (keys, s)
                add_to_summary(s, keys, msg)

        summary: dict = {}
        for vw in val_warns:
            assert issubclass(vw.category, ValidationWarning)
            msg = str(vw.message)
            if ": " in msg:
                keys_, *rest = msg.split(": ")
                msg = ": ".join(rest)
                keys = keys_.split(":")
            else:
                keys = []

            add_to_summary(summary, keys, msg)

        return summary


def get_format_version_module(type_: str, format_version: str):
    assert "." in format_version
    import bioimageio.spec

    type_ = get_spec_type_from_type(type_)
    version_mod_name = "v" + "_".join(format_version.split(".")[:2])
    try:
        return getattr(getattr(bioimageio.spec, type_), version_mod_name)
    except AttributeError:
        raise ValueError(
            f"Invalid RDF format version {format_version} for RDF type {type_}. "
            f"Submodule bioimageio.spec.{type_}.{version_mod_name} does not exist."
        )


def get_patched_format_version(type_: str, format_version: str):
    """return latest patched format version for given type and major/minor of format_version"""
    version_mod = get_format_version_module(type_, format_version)
    return version_mod.format_version


def get_spec_type_from_type(type_: Optional[str]):
    if type_ in ("model", "collection", "dataset"):
        return type_
    else:
        return "rdf"


def get_latest_format_version_module(type_: str):
    type_ = get_spec_type_from_type(type_)
    import bioimageio.spec

    try:
        return getattr(bioimageio.spec, type_)
    except AttributeError:
        raise ValueError(f"Invalid RDF type {type_}")


def get_latest_format_version(type_: str):
    return get_latest_format_version_module(type_).format_version


def get_class_name_from_type(type_: str):
    type_ = get_spec_type_from_type(type_)
    if type_ == "rdf":
        return "RDF"
    else:
        return type_.title()


def get_args_flat(tp):
    flat_args = []
    for a in get_args(tp):
        orig = get_origin(a)
        if orig is Literal or orig is Generic:
            flat_args += list(get_args(a))
        else:
            flat_args.append(a)

    return tuple(flat_args)


def nested_default_dict_as_nested_dict(nested_dd):
    if isinstance(nested_dd, dict):
        return {key: (nested_default_dict_as_nested_dict(value)) for key, value in nested_dd.items()}
    elif isinstance(nested_dd, list):
        return [nested_default_dict_as_nested_dict(value) for value in nested_dd]
    else:
        return nested_dd
